
import pytest

import platform
import os

import unittest.mock as mock

import pyDeltaRCM
from pyDeltaRCM.model import DeltaModel
# from pyDeltaRCM import shared_tools
from pyDeltaRCM import preprocessor

from . import utilities
# from .utilities import test_DeltaModel


class TestPreprocessorSingleJobSetups:

    # test high level python api
    def test_py_hlvl_wo_args(self):
        with pytest.raises(ValueError):
            _ = preprocessor.Preprocessor()

    def test_py_hlvl_runjobs_simple_param(self, tmp_path):
        # a single parameter
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'h0': 7.5})
        pp = preprocessor.Preprocessor(p)

        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        # a parameter that takes null as the default
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'hb': 7.5})
        pp = preprocessor.Preprocessor(p)

        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        # a parameter that takes null with null
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'hb': None})
        pp = preprocessor.Preprocessor(p)

        assert len(pp.file_list) == 1
        assert pp._is_completed is False

    def test_py_hlvl_tsteps_yml_runjobs_sngle(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'timesteps': 50})
        pp = preprocessor.Preprocessor(p)

        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        assert pp.config_dict['timesteps'] == 50
        assert not ('time' in pp.config_dict.keys())
        assert not ('time_years' in pp.config_dict.keys())

    def test_py_hlvl_time_yml_runjobs_sngle(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'time': 1000})
        pp = preprocessor.Preprocessor(p)

        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        assert pp.config_dict['time'] == 1000
        assert not ('timesteps' in pp.config_dict.keys())

    def test_py_hlvl_time_If_yml_runjobs_sngle(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'time': 10000,
                                      'If': 0.1})
        pp = preprocessor.Preprocessor(p)

        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        assert pp.config_dict['time'] == 10000
        assert pp.config_dict['If'] == 0.1
        assert not ('timesteps' in pp.config_dict.keys())

    def test_py_hlvl_timeyears_yml_runjobs_sngle(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'time_years': 3.16880878140e-05})
        pp = preprocessor.Preprocessor(p)

        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        assert pp.config_dict['time_years'] == pytest.approx(3.1688087e-05)
        assert not ('If' in pp.config_dict.keys())
        assert not ('time' in pp.config_dict.keys())
        assert not ('timesteps' in pp.config_dict.keys())

    def test_py_hlvl_timeyears_If_yml_runjobs_sngle(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'time_years': 0.00031688087814,
                                      'If': 0.1})
        pp = preprocessor.Preprocessor(p)

        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        assert pp.config_dict['time_years'] == pytest.approx(.00031688087814)
        assert pp.config_dict['If'] == 0.1
        assert not ('time' in pp.config_dict.keys())
        assert not ('timesteps' in pp.config_dict.keys())

    def test_py_hlvl_timesteps_and_yaml_argument(self, tmp_path):
        """
        Test that timesteps can come from cli, but others from yaml
        """
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_eta_figs': True})
        pp = preprocessor.Preprocessor(input_file=p, timesteps=2)

        assert type(pp.file_list) is list
        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        assert pp.config_dict['timesteps'] == 2
        assert (pp.config_dict['save_eta_figs'] is True)
        assert not ('time' in pp.config_dict.keys())

    def test_py_hlvl_timesteps_yaml_and_cli(self, tmp_path):
        """
        Test preference for command line argument
        """
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_eta_figs': True,
                                      'timesteps': 20})
        pp = preprocessor.Preprocessor(input_file=p, timesteps=13)

        assert type(pp.file_list) is list
        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        assert pp.config_dict['timesteps'] == 13
        assert (pp.config_dict['save_eta_figs'] is True)
        assert not ('time' in pp.config_dict.keys())

    def test_py_hlvl_timesteps_and_time_yaml_and_cli(self, tmp_path):
        """
        Test that preference is given to command line arguments

        note that time and timesteps both persist, precedence for these args
        is determined at runtime of run_jobs()
        """
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_eta_figs': True,
                                      'timesteps': 20,
                                      'time': 1000})
        pp = preprocessor.Preprocessor(input_file=p, timesteps=13, time=13000)

        assert type(pp.file_list) is list
        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        assert pp.config_dict['timesteps'] == 13
        assert pp.config_dict['time'] == 13000
        assert (pp.config_dict['save_eta_figs'] is True)
        assert ('time' in pp.config_dict.keys())
        assert ('timesteps' in pp.config_dict.keys())


class TestPreprocessorMatrixJobsSetups:

    def test_py_hlvl_mtrx_1list_timesteps_arg(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_matrix_to_file(f,
                                       ['f_bedload'],
                                       [[0.2, 0.6]])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p, timesteps=3)

        assert pp._has_matrix is True
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 2
        assert pp._is_completed is False

        f_bedload_list = pp._matrix_table[:, 1]

        assert sum([j == 0.2 for j in f_bedload_list]) == 1
        assert sum([j == 0.6 for j in f_bedload_list]) == 1
        assert pp.config_dict['timesteps'] == 3

    def test_py_hlvl_mtrx_1list_tsteps_cfg(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'timesteps', 3)
        utilities.write_matrix_to_file(f,
                                       ['f_bedload'],
                                       [[0.2, 0.6]])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p)

        assert pp._has_matrix is True
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 2
        assert pp._is_completed is False

        f_bedload_list = pp._matrix_table[:, 1]

        assert sum([j == 0.2 for j in f_bedload_list]) == 1
        assert sum([j == 0.6 for j in f_bedload_list]) == 1
        assert pp.config_dict['timesteps'] == 3

    def test_py_hlvl_mtrx_smthng_null(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'timesteps', 3)
        utilities.write_matrix_to_file(f,
                                       ['hb'],
                                       [[None, 2, 7]])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p)

        assert pp._has_matrix is True
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 3
        assert pp._is_completed is False

        hb_list = pp._matrix_table[:, 1]

        assert sum([j == 'None' for j in hb_list]) == 1
        assert sum([j == 2 for j in hb_list]) == 1
        assert sum([j == 7 for j in hb_list]) == 1

    def test_py_hlvl_mtrx_2list(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_matrix_to_file(f,
                                       ['f_bedload', 'u0'],
                                       [[0.2, 0.5, 0.6], [1.0, 1.5, 2.0]])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p)

        assert pp._has_matrix is True
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 9
        assert pp._is_completed is False

        f_bedload_list = pp._matrix_table[:, 1]
        u0_list = pp._matrix_table[:, 2]

        assert sum([j == 0.2 for j in f_bedload_list]) == 3
        assert sum([j == 0.5 for j in f_bedload_list]) == 3
        assert sum([j == 0.6 for j in f_bedload_list]) == 3
        assert sum([j == 1.0 for j in u0_list]) == 3
        assert sum([j == 1.5 for j in u0_list]) == 3
        assert sum([j == 2.0 for j in u0_list]) == 3

        comb_list = pp._matrix_table[:, 1:].tolist()
        assert [0.2, 2.0] in comb_list
        assert [0.5, 1.0] in comb_list
        assert not ([0.5, 0.2] in comb_list)

    def test_py_hlvl_mtrx_scientificnotation(self, tmp_path):
        """Test that preprocessor can read and write scinotation"""
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_matrix_to_file(f,
                                       ['f_bedload', 'SLR'],
                                       [[0.2, 0.5, 0.6], [0.00004, 1e-6]])
        f.close()

        # will convert everything to scientific notation
        pp = preprocessor.Preprocessor(input_file=p)

        SLR_list = pp._matrix_table[:, 2]
        assert sum([j == 4e-5 for j in SLR_list]) == 3
        assert sum([j == 0.000001 for j in SLR_list]) == 3

    def test_py_hlvL_mtrx_needs_out_dir(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        # missing out_dir in the config will throw an error
        # => utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_matrix_to_file(f,
                                       ['f_bedload'],
                                       [[0.2, 0.5, 0.6]])
        f.close()
        with pytest.raises(ValueError,
                           match=r'You must specify "out_dir" in YAML .*'):
            _ = preprocessor.Preprocessor(input_file=p)

    def test_py_hlvl_mtrx_outdir_exists_error(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_matrix_to_file(f,
                                       ['f_bedload'],
                                       [[0.2, 0.5, 0.6]])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p)

        # misc assertions
        assert pp._has_matrix is True

        # verify the directory exists
        assert pp._jobs_root == str(tmp_path / 'test')
        assert (tmp_path / 'test').is_dir()

        # try to create another preprocessor at same directory
        with pytest.raises(FileExistsError,
                           match=r'Job output directory .*'):
            _ = preprocessor.Preprocessor(input_file=p)

    def test_py_hlvl_mtrx_bad_len1(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        # bad configuration is a list of length 1
        utilities.write_matrix_to_file(f,
                                       ['f_bedload', 'u0'],
                                       [[0.2], [0.5, 0.6, 1.25]])
        f.close()
        with pytest.warns(UserWarning,
                          match=r'Length of matrix key "f_bedload" was 1,'):
            pp = preprocessor.Preprocessor(input_file=p)

        # check job length
        assert pp._has_matrix is True
        assert len(pp.file_list) == 3

    def test_py_hlvl_mtrx_bad_listinlist(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        # bad configuration will lead to error
        utilities.write_matrix_to_file(f,
                                       ['f_bedload', 'u0'],
                                       [[0.2, [0.5, 0.6]], [0.5, 1.25]])
        f.close()
        with pytest.raises(ValueError,
                           match=r'Depth of matrix expansion must not be > 1'):
            _ = preprocessor.Preprocessor(input_file=p)

    def test_py_hlvl_mtrx_bad_samekey(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'f_bedload', 0.3)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        # bad configuration will lead to error
        utilities.write_matrix_to_file(f,
                                       ['f_bedload', 'u0'],
                                       [[0.2, 0.5, 0.6], [0.5, 1.25]])
        f.close()
        with pytest.raises(ValueError,
                           match=r'You cannot specify the same key '
                                  'in the matrix .*'):  # noqa: E127
            _ = preprocessor.Preprocessor(input_file=p)

    def test_py_hlvl_mtrx_bad_colon(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        # bad configuration will lead to error
        utilities.write_matrix_to_file(f,
                                       ['f_bedload', 'u0:'],
                                       [[0.2, 0.5], [0.5, 1.25]])
        f.close()
        with pytest.raises(ValueError,
                           match=r'Colon operator found '
                                  'in matrix expansion key.'):  # noqa: E127
            _ = preprocessor.Preprocessor(input_file=p)

    def test_py_hlvl_mtrx_no_out_dir_in_mtrx(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_matrix_to_file(f,
                                       ['out_dir', 'f_bedload'],
                                       [['dir1', 'dir2'], [0.2, 0.5, 0.6]])
        f.close()
        with pytest.raises(ValueError,
                           match=r'You cannot specify "out_dir" as .*'):
            _ = preprocessor.Preprocessor(input_file=p, timesteps=3)

    def test_python_highlevelapi_matrix_verbosity(self, tmp_path, capsys):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'verbose', 1)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_matrix_to_file(f,
                                       ['f_bedload', 'u0'],
                                       [[0.2, 0.5, 0.6], [1.5, 2.0]])
        f.close()
        _ = preprocessor.Preprocessor(input_file=p, timesteps=3)

        captd = capsys.readouterr()
        assert 'Timestep: 0.0' not in captd.out
        assert 'Writing YAML file for job 0' in captd.out
        assert 'Writing YAML file for job 1' in captd.out
        assert 'Writing YAML file for job 2' in captd.out
        assert 'Writing YAML file for job 3' in captd.out
        assert 'Writing YAML file for job 4' in captd.out
        assert 'Writing YAML file for job 5' in captd.out
        assert 'Matrix expansion:' in captd.out
        assert '  dims 2' in captd.out
        assert '  jobs 6' in captd.out


class TestPreprocessorSetJobsSetups:

    def test_py_hlvl_set_two_sets(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_set_to_file(f, [{'f_bedload': 0.2, 'u0': 1},
                                        {'f_bedload': 0.4, 'u0': 2}])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p, timesteps=3)

        assert pp._has_set is True
        assert pp._has_matrix is False
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 2
        assert pp._is_completed is False

        f_bedload_list = pp._matrix_table[:, 1]

        assert sum([j == 0.2 for j in f_bedload_list]) == 1
        assert sum([j == 0.4 for j in f_bedload_list]) == 1
        assert pp.config_dict['timesteps'] == 3

    def test_py_hlvl_set_one_sets(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_set_to_file(f, [{'f_bedload': 0.77, 'u0': 1}])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p, timesteps=3)

        assert pp._has_set is True
        assert pp._has_matrix is False
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 1
        assert pp._is_completed is False

        f_bedload_list = pp._matrix_table[:, 1]

        assert sum([j == 0.77 for j in f_bedload_list]) == 1
        assert pp.config_dict['timesteps'] == 3

    def test_py_hlvl_set_two_mismatched_sets(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_set_to_file(f, [{'f_bedload': 0.2, 'u0': 1},
                                        {'h0': 0.4, 'u0': 2}])
        f.close()
        with pytest.raises(ValueError,
                           match=r'All keys in all sets *.'):  # noqa: E127
            _ = preprocessor.Preprocessor(input_file=p)

    def test_py_hlvl_set_bad_not_list(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        # utilities.write_set_to_file(f, [{'f_bedload': 0.77, 'u0': 1}])
        utilities.write_parameter_to_file(f, 'set', 'astring!')
        f.close()
        with pytest.raises(TypeError,
                           match=r'Set list must be *.'):
            _ = preprocessor.Preprocessor(input_file=p)

    def test_py_hlvl_set_bad_colon(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_set_to_file(f, [{'f_bedload:': 0.77, 'u0': 1}])
        f.close()
        with pytest.raises(ValueError,
                           match=r'Colon operator found '
                                  'in matrix expansion key.'):  # noqa: E127
            _ = preprocessor.Preprocessor(input_file=p)

    def test_output_table_correctly(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_set_to_file(
            f, [{'f_bedload': 0.77, 'u0': 1, 'C0_percent': 0.033}])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p, timesteps=3)

        # now check what the file looks like
        with open(os.path.join(
                tmp_path, 'test', 'jobs_parameters.txt'), 'r') as tbl:
            Lines = tbl.readlines()

        assert Lines[0] == 'job_id, f_bedload, u0, C0_percent\n'


class TestPreprocessorEnsembleJobsSetups:

    def test_py_hlvl_ensemble(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        f.close()
        pp = preprocessor.Preprocessor(input_file=p)
        # assertions for job creation
        assert pp._has_matrix is True
        assert pp._has_ensemble is True
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 2

    def test_py_hlvl_ensemble_1(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 1)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        f.close()
        with pytest.warns(UserWarning,
                          match=r'Ensemble was set to 1. *.'):
            pp = preprocessor.Preprocessor(input_file=p)

        # check that keys were reset and config set correctly
        assert pp._has_ensemble is False
        assert pp._has_matrix is False
        assert len(pp.file_list) == 1

    def test_py_hlvl_ensemble_badtype(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2.0)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        f.close()
        with pytest.raises(TypeError,
                           match=r'Invalid ensemble type, '
                                  'must be an integer.'):  # noqa: E127
            _ = preprocessor.Preprocessor(input_file=p)

    def test_py_hlvl_ensemble_double_seeds(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'seed', 1)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        f.close()
        with pytest.raises(ValueError,
                           match=r'You cannot specify the same key in the '
                                  'matrix configuration and '  # noqa: E127
                                  'fixed configuration. Key "seed" was '
                                  'specified in both.'):
            _ = preprocessor.Preprocessor(input_file=p)


class TestPreprocessorMatrixAndEnsembleJobsSetups:

    def test_py_hlvl_ensemble_with_matrix(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_matrix_to_file(f,
                                       ['f_bedload'],
                                       [[0.2, 0.5, 0.6]])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p)
        # assertions for job creation
        assert pp._has_matrix is True
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 6

    def test_py_hlvl_ensemble_matrix_seeds(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_matrix_to_file(f,
                                       ['seed'],
                                       [[1, 2]])
        f.close()
        with pytest.raises(ValueError,
                           match=r'Random seeds cannot be specified in '
                                  'the matrix, if an "ensemble" number '  # noqa: E127, E501
                                  'is specified as well.'):
            _ = preprocessor.Preprocessor(input_file=p, timesteps=3)


class TestPreprocessorSetAndEnsembleJobsSetups:

    def test_1_set_ensemble_3(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'ensemble', 3)
        utilities.write_set_to_file(f, [{'f_bedload': 0.77, 'u0': 1}])
        f.close()
        pp = preprocessor.Preprocessor(input_file=p, timesteps=3)

        assert pp._has_set is True
        assert pp._has_matrix is False
        assert pp._has_ensemble is True
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 3
        assert pp._is_completed is False

        f_bedload_list = pp._matrix_table[:, 1]

        assert sum([j == 0.77 for j in f_bedload_list]) == 3
        assert pp.config_dict['timesteps'] == 3


class TestPreprocessorSetAndMatrixJobsSetups:

    def test_not_possible(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_set_to_file(f, [{'f_bedload': 0.77, 'h0': 1}])
        utilities.write_matrix_to_file(f,
                                       ['u0'],
                                       [[0.2, 0.5, 0.6]])
        f.close()

        with pytest.raises(ValueError,
                           match=r'Cannot use "matrix" and "set" .*'):
            _ = preprocessor.Preprocessor(input_file=p, timesteps=3)


class TestPreprocessorParallelJobsSetups:
    """
    Note that parallel job setup will work on any operating system. This is
    helpful for someone who wants to use the preprocessor to tinker with job
    configurations locally, but will run jobs on a remote Linux cluster/HPC.
    """

    def test_py_hlvl_parallel_works_onejob(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'parallel', True)
        f.close()
        pp = preprocessor.Preprocessor(input_file=p)
        # assertions for job creation
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 1
        assert pp._has_matrix is False

        assert pp.config_dict['parallel'] is True

    def test_py_hlvl_parallel_boolean_yaml(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'parallel', True)
        f.close()
        pp = preprocessor.Preprocessor(input_file=p)
        # assertions for job creation
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 2
        assert pp._has_matrix is True

        assert pp.config_dict['parallel'] is True

    def test_py_hlvl_parallel_boolean_cli(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        f.close()
        pp = preprocessor.Preprocessor(input_file=p, parallel=True)
        # assertions for job creation
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 2
        assert pp._has_matrix is True

        assert pp.config_dict['parallel'] is True

    def test_py_hlvl_parallel_integer_yaml(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        utilities.write_parameter_to_file(f, 'parallel', 2)
        f.close()
        pp = preprocessor.Preprocessor(input_file=p)
        # assertions for job creation
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 2
        assert pp._has_matrix is True

        assert pp.config_dict['parallel'] == 2

    def test_py_hlvl_parallel_integer_cli(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'ensemble', 2)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'test')
        f.close()
        pp = preprocessor.Preprocessor(input_file=p, parallel=2)
        # assertions for job creation
        assert type(pp.file_list) is list
        assert len(pp.file_list) == 2
        assert pp._has_matrix is True
        assert pp._is_completed is False

        assert pp.config_dict['parallel'] == 2


class TestPreprocessorRunJobs:

    def test_dryrun(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        pp = preprocessor.Preprocessor(p)

        pp._dryrun = True

        # run the method
        pp.run_jobs()

        assert pp._is_completed is False

    def test_run_single_serial_job(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        pp = preprocessor.Preprocessor(p)

        # patch jobs
        with mock.patch('pyDeltaRCM.preprocessor._SerialJob') as ptch:

            # run the method
            pp.run_jobs()

        assert ptch.call_count == 1
        assert len(pp.job_list) == 1
        assert pp._is_completed is True

    def test_run_two_serial_jobs(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        pp = preprocessor.Preprocessor(p)

        pp._file_list = [*pp.file_list] * 2
        pp._config_list = [*pp.config_list] * 2

        # patch jobs
        with mock.patch('pyDeltaRCM.preprocessor._SerialJob') as ptch:

            # run the method
            pp.run_jobs()

        assert ptch.call_count == 2
        assert len(pp.job_list) == 2
        assert pp._is_completed is True

    @pytest.mark.skipif(
        platform.system() != 'Linux',
        reason='Parallel support only on Linux OS.')
    def test_run_five_parallel_jobs_bool(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        pp = preprocessor.Preprocessor(p, parallel=True)

        # expand the list to include multiples
        pp._file_list = [*pp.file_list] * 5
        pp._config_list = [*pp.config_list] * 5

        # patch jobs and semaphore to check number of processes called
        with mock.patch('pyDeltaRCM.preprocessor._ParallelJob') as ptch, \
             mock.patch('multiprocessing.Semaphore') as sem:

            # run the method
            pp.run_jobs()

        # assertions
        assert ptch.call_count == 5
        # can only assert sem was called:
        #   number called with will depend on number of cpu cores)
        sem.assert_called()
        assert len(pp.job_list) == 5
        assert pp._is_completed is True

    @pytest.mark.skipif(
        platform.system() != 'Linux',
        reason='Parallel support only on Linux OS.')
    def test_run_five_parallel_jobs_int_two(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        pp = preprocessor.Preprocessor(p, parallel=2)

        # expand the list to include multiples
        pp._file_list = [*pp.file_list] * 5
        pp._config_list = [*pp.config_list] * 5

        # patch jobs and semaphore to check number of processes called
        with mock.patch('pyDeltaRCM.preprocessor._ParallelJob') as ptch, \
             mock.patch('multiprocessing.Semaphore') as sem:

            # run the method
            pp.run_jobs()

        # assertions
        assert ptch.call_count == 5
        sem.assert_called_with(2)
        assert len(pp.job_list) == 5
        assert pp._is_completed is True

    @pytest.mark.skipif(
        platform.system() != 'Linux',
        reason='Parallel support only on Linux OS.')
    def test_run_five_parallel_jobs_bad_types(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        pp_float = preprocessor.Preprocessor(p, parallel=3.33)
        pp_str = preprocessor.Preprocessor(p, parallel='string!')

        # patch jobs and semaphore to check number of processes called
        with mock.patch('pyDeltaRCM.preprocessor._ParallelJob'), \
             mock.patch('multiprocessing.Semaphore'):

            with pytest.raises(ValueError, match=r'Parallel flag *.'):
                # run the method
                pp_float.run_jobs()

        with mock.patch('pyDeltaRCM.preprocessor._ParallelJob'), \
             mock.patch('multiprocessing.Semaphore'):

            with pytest.raises(ValueError, match=r'Parallel flag *.'):
                # run the method
                pp_str.run_jobs()

    @pytest.mark.skipif(
        platform.system() == 'Linux',
        reason='Parallel support only on Linux OS.')
    def test_run_parallel_notimplemented_nonlinux(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        pp = preprocessor.Preprocessor(p, parallel=True)

        # patch jobs and semaphore to check number of processes called
        with pytest.raises(NotImplementedError):

            # run the method
            pp.run_jobs()


@mock.patch.multiple(pyDeltaRCM.preprocessor._BaseJob,
                     __abstractmethods__=set())
class TestBaseJob:
    """
    To test the BaseJob, we patch the base job with a filled abstract method
    `.run()`.

    .. note:: This patch is handled at the class level above!!
    """

    def test_wo_timesteps(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        with pytest.raises(ValueError,
                           match=r'You must specify a run duration *.'):
            _ = preprocessor._BaseJob(
                i=0, input_file=p,
                config_dict={})

    def test_timeargs_precedence_tstepsovertime(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        basej = preprocessor._BaseJob(
            i=0, input_file=p,
            config_dict={'time': 10000,
                         'timesteps': 2})

        assert isinstance(basej.deltamodel, DeltaModel)
        assert basej._job_end_time == basej.deltamodel._dt * 2
        assert basej._job_end_time != 10000

    def test_timeargs_precedence_tstepsovertimeyears(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        basej = preprocessor._BaseJob(
            i=0, input_file=p,
            config_dict={'time_years': 10000,
                         'timesteps': 2})

        assert isinstance(basej.deltamodel, DeltaModel)
        assert basej._job_end_time == basej.deltamodel._dt * 2
        assert basej._job_end_time != 10000
        assert basej._job_end_time != 10000 * 365.25

    def test_timeargs_precedence_timeovertimeyears(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        basej = preprocessor._BaseJob(
            i=0, input_file=p,
            config_dict={'time_years': 10000,
                         'time': 900})

        assert isinstance(basej.deltamodel, DeltaModel)
        assert basej._job_end_time == 900
        assert basej._job_end_time != 10000
        assert basej._job_end_time != 10000 * 365.25


class TestSerialJob:

    def test_run(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        sj = preprocessor._SerialJob(
            i=0, input_file=p,
            config_dict={'timesteps': 10})

        # modify the save interval to be twice dt
        sj.deltamodel._save_dt = 2 * sj.deltamodel._dt
        sj.deltamodel._checkpoint_dt = 2 * sj.deltamodel._dt

        # mock top-level methods, verify call was made to each
        sj.deltamodel.solve_water_and_sediment_timestep = mock.MagicMock()
        sj.deltamodel.apply_subsidence = mock.MagicMock()
        sj.deltamodel.finalize_timestep = mock.MagicMock()
        sj.deltamodel.log_model_time = mock.MagicMock()
        sj.deltamodel.logger = mock.MagicMock()
        sj.deltamodel.output_data = mock.MagicMock()
        sj.deltamodel.output_checkpoint = mock.MagicMock()
        sj.deltamodel.finalize = mock.MagicMock()

        # run the method (call update 10 times)
        sj.run()

        # assertions
        assert sj.deltamodel._time_iter == 10
        assert sj.deltamodel.output_data.call_count == 10
        assert sj.deltamodel.solve_water_and_sediment_timestep.call_count == 10
        assert sj.deltamodel.apply_subsidence.call_count == 10
        assert sj.deltamodel.finalize_timestep.call_count == 10
        assert sj.deltamodel.log_model_time.call_count == 10
        assert sj.deltamodel.output_checkpoint.call_count == 10
        assert sj.deltamodel.finalize.call_count == 1

        # check the log outputs for successes
        _calls = [mock.call('job: 0, stage: 1, code: 0'),
                  mock.call('job: 0, stage: 2, code: 0')]
        sj.deltamodel.logger.info.assert_has_calls(_calls)

    def test_run_error_update(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        sj = preprocessor._SerialJob(
            i=99, input_file=p,
            config_dict={'timesteps': 10})

        # modify the save interval to be twice dt
        sj.deltamodel._save_dt = 2 * sj.deltamodel._dt
        sj.deltamodel._checkpoint_dt = 2 * sj.deltamodel._dt

        # mock top-level methods, verify call was made to each
        sj.deltamodel.solve_water_and_sediment_timestep = mock.MagicMock()
        sj.deltamodel.apply_subsidence = mock.MagicMock()
        sj.deltamodel.finalize_timestep = mock.MagicMock()
        sj.deltamodel.log_model_time = mock.MagicMock()
        sj.deltamodel.logger = mock.MagicMock()
        sj.deltamodel.output_data = mock.MagicMock()
        sj.deltamodel.output_checkpoint = mock.MagicMock(
            side_effect=RuntimeError('error!'))
        sj.deltamodel.finalize = mock.MagicMock()

        # run the method (error on `output_checkpoint`)
        sj.run()

        # assertions
        assert sj.deltamodel._time_iter == 1
        assert sj.deltamodel.solve_water_and_sediment_timestep.call_count == 1
        assert sj.deltamodel.apply_subsidence.call_count == 1
        assert sj.deltamodel.finalize_timestep.call_count == 1
        assert sj.deltamodel.log_model_time.call_count == 1
        assert sj.deltamodel.output_data.call_count == 1
        assert sj.deltamodel.output_checkpoint.call_count == 1
        assert sj.deltamodel.finalize.call_count == 1

        # check the log outputs for success/failure
        _info_calls = [mock.call('job: 99, stage: 2, code: 0')]
        _error_calls = [mock.call('job: 99, stage: 1, code: 1, msg: error!')]
        sj.deltamodel.logger.info.assert_has_calls(_info_calls)
        sj.deltamodel.logger.error.assert_has_calls(_error_calls)
        sj.deltamodel.logger.exception.assert_called_once()

    def test_run_error_finalize(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        sj = preprocessor._SerialJob(
            i=0, input_file=p,
            config_dict={'timesteps': 10})

        # modify the save interval to be twice dt
        sj.deltamodel._save_dt = 2 * sj.deltamodel._dt
        sj.deltamodel._checkpoint_dt = 2 * sj.deltamodel._dt

        # mock top-level methods, verify call was made to each
        sj.deltamodel.solve_water_and_sediment_timestep = mock.MagicMock()
        sj.deltamodel.apply_subsidence = mock.MagicMock()
        sj.deltamodel.finalize_timestep = mock.MagicMock()
        sj.deltamodel.log_model_time = mock.MagicMock()
        sj.deltamodel.logger = mock.MagicMock()
        sj.deltamodel.output_data = mock.MagicMock()
        sj.deltamodel.output_checkpoint = mock.MagicMock()
        sj.deltamodel.finalize = mock.MagicMock(
            side_effect=RuntimeError('error!'))

        # run the method (error on `output_checkpoint`)
        sj.run()

        # assertions
        assert sj.deltamodel._time_iter == 10
        assert sj.deltamodel.solve_water_and_sediment_timestep.call_count == 10
        assert sj.deltamodel.apply_subsidence.call_count == 10
        assert sj.deltamodel.finalize_timestep.call_count == 10
        assert sj.deltamodel.log_model_time.call_count == 10
        assert sj.deltamodel.output_data.call_count == 10
        assert sj.deltamodel.output_checkpoint.call_count == 10
        assert sj.deltamodel.finalize.call_count == 1

        # check the log outputs for success/failure
        _info_calls = [mock.call('job: 0, stage: 1, code: 0')]
        _error_calls = [mock.call('job: 0, stage: 2, code: 1, msg: error!')]
        sj.deltamodel.logger.info.assert_has_calls(_info_calls)
        sj.deltamodel.logger.error.assert_has_calls(_error_calls)
        sj.deltamodel.logger.exception.assert_called_once()


class TestParallelJob:

    def test_run(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        Q = mock.MagicMock()
        S = mock.MagicMock()

        pj = preprocessor._ParallelJob(
            i=0, queue=Q, sema=S, input_file=p,
            config_dict={'timesteps': 10})

        # modify the save interval to be twice dt
        pj.deltamodel._save_dt = 2 * pj.deltamodel._dt
        pj.deltamodel._checkpoint_dt = 2 * pj.deltamodel._dt

        # mock top-level methods, verify call was made to each
        pj.deltamodel.solve_water_and_sediment_timestep = mock.MagicMock()
        pj.deltamodel.apply_subsidence = mock.MagicMock()
        pj.deltamodel.finalize_timestep = mock.MagicMock()
        pj.deltamodel.log_model_time = mock.MagicMock()
        pj.deltamodel.logger = mock.MagicMock()
        pj.deltamodel.output_data = mock.MagicMock()
        pj.deltamodel.output_checkpoint = mock.MagicMock()
        pj.deltamodel.finalize = mock.MagicMock()

        # mock initialization and their hooks
        pj.deltamodel.init_output_file = mock.MagicMock()
        pj.deltamodel.hook_init_output_file = mock.MagicMock()
        pj.deltamodel.hook_output_data = mock.MagicMock()
        pj.deltamodel.hook_output_checkpoint = mock.MagicMock()

        # run the method (call update 10 times)
        pj.run()

        # assertions
        assert pj.deltamodel._time_iter == 10
        assert pj.deltamodel.solve_water_and_sediment_timestep.call_count == 10
        assert pj.deltamodel.apply_subsidence.call_count == 10
        assert pj.deltamodel.finalize_timestep.call_count == 10
        assert pj.deltamodel.log_model_time.call_count == 10
        assert pj.deltamodel.output_data.call_count == 11  # once in init
        assert pj.deltamodel.output_checkpoint.call_count == 11  # once in init
        assert pj.deltamodel.finalize.call_count == 1
        # hooks/initialization
        assert pj.deltamodel.init_output_file.call_count == 1
        assert pj.deltamodel.hook_init_output_file.call_count == 1
        assert pj.deltamodel.hook_output_data.call_count == 11
        assert pj.deltamodel.hook_output_checkpoint.call_count == 11

        # check the log outputs for successes
        _calls = [mock.call({'job': 0, 'stage': 0, 'code': 0}),
                  mock.call({'job': 0, 'stage': 1, 'code': 0}),
                  mock.call({'job': 0, 'stage': 2, 'code': 0})]
        Q.put.assert_has_calls(_calls)

    def test_run_error_update(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        Q = mock.MagicMock()
        S = mock.MagicMock()

        pj = preprocessor._ParallelJob(
            i=99, queue=Q, sema=S, input_file=p,
            config_dict={'timesteps': 10})

        # modify the save interval to be twice dt
        pj.deltamodel._save_dt = 2 * pj.deltamodel._dt
        pj.deltamodel._checkpoint_dt = 2 * pj.deltamodel._dt

        # mock top-level methods, verify call was made to each
        pj.deltamodel.solve_water_and_sediment_timestep = mock.MagicMock()
        pj.deltamodel.apply_subsidence = mock.MagicMock()
        pj.deltamodel.finalize_timestep = mock.MagicMock()
        pj.deltamodel.log_model_time = mock.MagicMock(
            side_effect=RuntimeError('error!'))
        pj.deltamodel.logger = mock.MagicMock()
        pj.deltamodel.output_data = mock.MagicMock()
        pj.deltamodel.output_checkpoint = mock.MagicMock()
        pj.deltamodel.finalize = mock.MagicMock()

        # run the method (error on `output_checkpoint`)
        pj.run()

        # assertions
        assert pj.deltamodel._time_iter == 1
        assert pj.deltamodel.solve_water_and_sediment_timestep.call_count == 1
        assert pj.deltamodel.apply_subsidence.call_count == 1
        assert pj.deltamodel.finalize_timestep.call_count == 1
        assert pj.deltamodel.log_model_time.call_count == 1
        assert pj.deltamodel.output_data.call_count == 1
        assert pj.deltamodel.output_checkpoint.call_count == 1
        assert pj.deltamodel.finalize.call_count == 1

        # check the log outputs for success/failure
        _info_calls = [mock.call({'job': 99, 'stage': 0, 'code': 0}),
                       mock.call({'job': 99, 'stage': 2, 'code': 0})]
        _error_calls = [mock.call({'job': 99, 'stage': 1,
                                   'code': 1, 'msg': 'error!'})]
        Q.put.assert_has_calls(_info_calls, any_order=True)
        Q.put.assert_has_calls(_error_calls)
        pj.deltamodel.logger.error.assert_has_calls(
            [mock.call('error!')])
        pj.deltamodel.logger.exception.assert_called_once()

    def test_run_error_finalize(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')

        Q = mock.MagicMock()
        S = mock.MagicMock()

        pj = preprocessor._ParallelJob(
            i=0, queue=Q, sema=S, input_file=p,
            config_dict={'timesteps': 10})

        # modify the save interval to be twice dt
        pj.deltamodel._save_dt = 2 * pj.deltamodel._dt
        pj.deltamodel._checkpoint_dt = 2 * pj.deltamodel._dt

        # mock top-level methods, verify call was made to each
        pj.deltamodel.solve_water_and_sediment_timestep = mock.MagicMock()
        pj.deltamodel.apply_subsidence = mock.MagicMock()
        pj.deltamodel.finalize_timestep = mock.MagicMock()
        pj.deltamodel.log_model_time = mock.MagicMock()
        pj.deltamodel.logger = mock.MagicMock()
        pj.deltamodel.output_data = mock.MagicMock()
        pj.deltamodel.output_checkpoint = mock.MagicMock()
        pj.deltamodel.finalize = mock.MagicMock(
            side_effect=RuntimeError('error!'))

        # run the method (error on `output_checkpoint`)
        pj.run()

        # assertions
        assert pj.deltamodel._time_iter == 10
        assert pj.deltamodel.solve_water_and_sediment_timestep.call_count == 10
        assert pj.deltamodel.apply_subsidence.call_count == 10
        assert pj.deltamodel.finalize_timestep.call_count == 10
        assert pj.deltamodel.log_model_time.call_count == 10
        assert pj.deltamodel.output_data.call_count == 11  # once in init
        assert pj.deltamodel.output_checkpoint.call_count == 11  # once in init
        assert pj.deltamodel.finalize.call_count == 1

        # check the log outputs for success/failure
        _info_calls = [mock.call({'job': 0, 'stage': 0, 'code': 0}),
                       mock.call({'job': 0, 'stage': 1, 'code': 0})]
        _error_calls = [mock.call({'job': 0, 'stage': 2,
                                   'code': 1, 'msg': 'error!'})]
        Q.put.assert_has_calls(_info_calls, any_order=True)
        Q.put.assert_has_calls(_error_calls)
        pj.deltamodel.logger.error.assert_has_calls(
            [mock.call('error!')])
        pj.deltamodel.logger.exception.assert_called_once()


class TestWriteYamlConfigToFile:

    def test_write_single_int(self, tmp_path):
        # set up what to write
        file_path = tmp_path / 'output.yml'
        yaml_dict = {'variable': 1}

        # write the file
        preprocessor._write_yaml_config_to_file(yaml_dict, file_path)

        with open(file_path) as f:
            _returned = ' '.join(f.readlines())

        assert _returned == 'variable: 1\n'

    def test_write_single_float(self, tmp_path):
        # set up what to write
        file_path = tmp_path / 'output.yml'
        yaml_dict = {'variable': 1.5}

        # write the file
        preprocessor._write_yaml_config_to_file(yaml_dict, file_path)

        with open(file_path) as f:
            _returned = ' '.join(f.readlines())

        assert _returned == 'variable: 1.5\n'

    def test_write_single_string(self, tmp_path):
        # set up what to write
        file_path = tmp_path / 'output.yml'
        yaml_dict = {'variable': 'a string'}

        # write the file
        preprocessor._write_yaml_config_to_file(yaml_dict, file_path)

        with open(file_path) as f:
            _returned = ' '.join(f.readlines())

        assert _returned == 'variable: a string\n'

    def test_write_single_bool(self, tmp_path):
        # set up what to write
        file_path = tmp_path / 'output.yml'
        yaml_dict = {'variable': False}

        # write the file
        preprocessor._write_yaml_config_to_file(yaml_dict, file_path)

        with open(file_path) as f:
            _returned = ' '.join(f.readlines())

        assert _returned == 'variable: False\n'

    def test_write_single_null(self, tmp_path):
        # set up what to write
        file_path = tmp_path / 'output.yml'
        yaml_dict = {'variable': None}

        # write the file
        preprocessor._write_yaml_config_to_file(yaml_dict, file_path)

        with open(file_path) as f:
            _returned = ' '.join(f.readlines())

        assert _returned == 'variable: null\n'

    def test_write_multiple_one_each(self, tmp_path):
        # set up what to write
        file_path = tmp_path / 'output.yml'
        yaml_dict = {'variable1': 1,
                     'variable2': 2.2,
                     'variable3': 'a string',
                     'variable4': False,
                     'variable5': None}

        # write the file
        preprocessor._write_yaml_config_to_file(yaml_dict, file_path)

        with open(file_path) as f:
            _returned = ' '.join(f.readlines())
        print(_returned)

        assert _returned == ('variable1: 1\n variable2: 2.2\n '
                             'variable3: a string\n variable4: False\n '
                             'variable5: null\n')


class TestPreprocessorImported:

    def test_Preprocessor_toplevelimport(self):
        import pyDeltaRCM

        assert 'Preprocessor' in dir(pyDeltaRCM)
        assert pyDeltaRCM.Preprocessor is pyDeltaRCM.preprocessor.Preprocessor


class TestScaleRelativeSeaLeveLRiseRate():

    def test_scale_If_1(self):
        scaled = preprocessor.scale_relative_sea_level_rise_rate(5, If=1)
        assert scaled == 5 / 1000 / 365.25 / 86400

    def test_scale_If_0p1(self):
        scaled = preprocessor.scale_relative_sea_level_rise_rate(5, If=0.1)
        assert scaled == 5 / 1000 / 365.25 / 86400 * 10



#! /usr/bin/env python

import os
import warnings

import numpy as np
import matplotlib.pyplot as plt
import mpl_toolkits.axes_grid1 as axtk

from scipy.sparse import lil_matrix, csc_matrix, hstack

import abc

from . import shared_tools


class iteration_tools(abc.ABC):
    """Tools relating to the updating of the model and model I/O.

    Tools defined in this class include steps to iterate for one timestep,
    finalize timesteps, and saving output figures, grids, and checkpoints.
    Additionally, most stratigraphy-related operations are defined here, since
    these operations largely occur when saving and updating the model.
    """

    def solve_water_and_sediment_timestep(self):
        """Run water and sediment operations for one timestep.

        The first operation called by :meth:`update`, this method iterates the
        water surface calculation and sediment parcel routing routines.

        Parameters
        ----------

        Returns
        -------
        """
        # start the model operations
        self.eta0 = np.copy(self.eta)  # copy

        # water iterations
        self.hook_route_water()
        self.route_water()
        self.hook_after_route_water()

        # sediment iteration
        self.hook_route_sediment()
        self.route_sediment()
        self.hook_after_route_sediment()

    def run_one_timestep(self):
        """Deprecated, since v1.3.1. Use :obj:`solve_water_and_sediment_timestep`."""
        _msg = ('`run_one_timestep` and `hook_run_one_timestep` are '
                'deprecated and have been replaced with '
                '`solve_water_and_sediment_timestep`. '
                'Running `solve_water_and_sediment_timestep` now, but '
                'this will be removed in future release.')
        self.logger.warning(_msg)
        warnings.warn(UserWarning(_msg))
        self.solve_water_and_sediment_timestep()

    def apply_subsidence(self):
        """Apply subsidence pattern.

        Apply subsidence to domain if toggle_subsidence is True, and
        :obj:`~pyDeltaRCM.DeltaModel.time` is ``>=``
        :obj:`~pyDeltaRCM.DeltaModel.start_subsidence`. Note, that the
        configuration of the :obj:`~pyDeltaRCM.DeltaModel.update()` method
        determines that the subsidence may be applied before the model time
        is incremented, such that subsidence will begin on the step
        *following* the time step that brings the model to ``time ==
        start_subsidence``.

        Parameters
        ----------

        Returns
        -------

        """
        if self._toggle_subsidence:

            if self._time >= self._start_subsidence:

                _msg = 'Applying subsidence'
                self.log_info(_msg, verbosity=1)

                self.eta[:] = self.eta - self.sigma

    def finalize_timestep(self):
        """Finalize timestep.

        Clean up after sediment routing. This includes a correction for
        flooded cells that are not "wet" (via :meth:`flooding_correction`).

        Update sea level if baselevel changes between timesteps.

        Parameters
        ----------

        Returns
        -------

        """
        _msg = 'Finalizing timestep'
        self.log_info(_msg, verbosity=2)

        self.flooding_correction()
        self.stage[:] = np.maximum(self.stage, self._H_SL)
        self.depth[:] = np.maximum(self.stage - self.eta, 0)

        self.eta[0, self.inlet] = self.stage[0, self.inlet] - self._h0
        self.depth[0, self.inlet] = self._h0

        self.hook_compute_sand_frac()
        self.compute_sand_frac()

        self.H_SL = self._H_SL + self._SLR * self._dt

    def log_info(self, message, verbosity=0):
        """Log message dependent on verbosity settings.

        Parameters
        ----------
        message : :obj:`str`
            Message string to write to the log as info.

        verbosity : :obj:`int`, optional
            Verbosity threshold, whether to write the message to the log or
            not. Default value is `0`, or i.e. to always log.
        """
        if self._verbose >= verbosity:
            self.logger.info(message)

    def log_model_time(self):
        """Log the time of the model.

        Reports the time to the log file, and depending on verbosity, will
        report it to stdout.
        """
        _timemsg = 'Time: {time:.{digits}f}; timestep: {timestep:g}'.format(
            time=self._time, timestep=self._time_iter, digits=1)
        self.logger.info(_timemsg)
        if self._verbose > 0:
            print(_timemsg)

    def output_data(self):
        """Output grids and figures if needed.

        """
        if self._save_time_since_data >= self.save_dt:

            self.save_grids_and_figs()

            self._save_iter += int(1)
            self._save_time_since_data = 0

    def output_checkpoint(self):
        """Output checkpoint if needed.

        Save checkpoint data (including rng state) so that the model can be
        resumed from this time.

        Parameters
        ----------

        Returns
        -------

        """
        if self._save_time_since_checkpoint >= self.checkpoint_dt:

            if self._save_checkpoint:

                _msg = 'Saving checkpoint'
                self.log_info(_msg, verbosity=1)

                self.save_the_checkpoint()

                if self._checkpoint_dt != self._save_dt:
                    _msg = ('Grid save interval and checkpoint interval are '
                            'not identical, this may result in duplicate '
                            'entries in the output NetCDF4 after resuming '
                            'the model run.')
                    self.logger.warning(_msg)

                self._save_time_since_checkpoint = 0

    def compute_sand_frac(self):
        """Compute the sand fraction as a continous updating data field.

        Parameters
        ----------

        Returns
        -------

        """
        _msg = 'Computing bed sand fraction'
        self.log_info(_msg, verbosity=2)

        # layer attributes at time t
        actlyr_thick = self._active_layer_thickness
        actlyr_top = np.copy(self.eta0)
        actlyr_bot = actlyr_top - actlyr_thick

        deta = self.eta - self.eta0

        # everywhere the bed has degraded this timestep
        whr_deg = (deta < 0)
        if np.any(whr_deg):
            # find where the erosion exceeded the active layer
            whr_unkwn = self.eta < actlyr_bot

            # update sand_frac in unknown to the boundary condition
            self.sand_frac[whr_unkwn] = self._sand_frac_bc

            # find where erosion was into active layer
            whr_actero = np.logical_and(whr_deg, self.eta >= actlyr_bot)

            # update sand_frac to active_layer value
            self.sand_frac[whr_actero] = self.active_layer[whr_actero]

        # handle aggradation/deposition
        whr_agg = (deta > 0)
        whr_agg = np.logical_or(
            (self.Vp_dep_sand > 0), (self.Vp_dep_mud > 0.000001))
        if np.any(whr_agg):
            # sand_frac and active_layer becomes the mixture of the deposit
            mixture = (self.Vp_dep_sand[whr_agg] /
                       (self.Vp_dep_mud[whr_agg] +
                        self.Vp_dep_sand[whr_agg]))

            # update sand_frac in act layer to this value
            self.sand_frac[whr_agg] = mixture
            self.active_layer[whr_agg] = mixture

    def save_grids_and_figs(self):
        """Save grids and figures.

        Save grids and/or plots of specified variables (``eta``, `discharge``,
        ``velocity``, ``depth``, and ``stage``, depending on configuration of
        the relevant flags in the YAML configuration file.

        .. note:

            This method is called often throughout the model, each
            occurance :obj:`save_dt` is elapsed in model time.

        Parameters
        ----------

        Returns
        -------

        """
        save_idx = self.save_iter

        _msg = ' '.join((
            'Saving data to output file:',
            str(save_idx).zfill(5)))
        self.log_info(_msg, verbosity=1)

        if (self._save_metadata or self._save_any_grids):
            self.output_netcdf.variables['time'][save_idx] = self._time

        # ------------------ Figures ------------------
        if len(self._save_fig_list) > 0:

            _msg = 'Saving figures'
            self.log_info(_msg, verbosity=2)

            for f in self._save_fig_list.keys():
                _attr = getattr(self, self._save_fig_list[f][0])
                if isinstance(_attr, np.ndarray):
                    if _attr.shape == (self.L, self.W):
                        _fig = self.make_figure(self._save_fig_list[f][0],
                                                self._time)
                        self.save_figure(_fig, directory=self.prefix,
                                         filename_root=f+'_',
                                         timestep=self.save_iter)
                    else:
                        raise AttributeError('Attribute "{_k}" is not of the '
                                             'right shape to be saved as a '
                                             'figure using the built-in '
                                             'methods. Expected a shape of '
                                             '"{_expshp}", but it has a shape '
                                             'of "{_wasshp}". Consider making '
                                             'a custom plotting utility to '
                                             'visualize this attribute.'
                                             .format(_k=f,
                                                     _expshp=(self.L, self.W),
                                                     _wasshp=_attr.shape))
                else:
                    raise AttributeError('Only plotting of np.ndarray-type '
                                         'attributes is natively supported. '
                                         'Input "{_k}" was of type "{_wt}".'
                                         .format(_k=f, _wt=type(_attr)))

        # ------------------ grids ------------------
        if self._save_any_grids:

            _msg = 'Saving grids'
            self.log_info(_msg, verbosity=2)

            _var_list = list(self._save_var_list.keys())
            _var_list.remove('meta')
            for _val in _var_list:
                self.save_grids(_val, getattr(self,
                                              self._save_var_list[_val][0]),
                                save_idx)

        # ------------------ metadata ------------------
        if self._save_metadata:

            _msg = 'Saving metadata'
            self.log_info(_msg, verbosity=2)

            for _val in self._save_var_list['meta'].keys():
                # use knowledge of time-varying values to save them
                if (self._save_var_list['meta'][_val][0] is None):
                    self.output_netcdf['meta'][_val][save_idx] = \
                        getattr(self, _val)

        # -------------------- sync --------------------
        if (self._save_metadata or self._save_any_grids):

            _msg = 'Syncing data to output file'
            self.log_info(_msg, verbosity=2)

            self.output_netcdf.sync()

    def make_figure(self, var, timestep):
        """Create a figure.

        Parameters
        ----------
        var : :obj:`str`
            Which variable to plot into the figure. Specified as a string and
            looked up via `getattr`.

        Returns
        -------
        fig : :obj:`matplotlib.figure`
            The created figure object.
        """
        _data = getattr(self, var)

        fig, ax = plt.subplots()
        im = ax.pcolormesh(self.X, self.Y, _data, shading='flat')
        ax.set_xlim((0, self._Width))
        ax.set_ylim((0, self._Length))
        ax.set_aspect('equal', adjustable='box')
        divider = axtk.axes_divider.make_axes_locatable(ax)
        cax = divider.append_axes("right", size="2%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.ax.tick_params(labelsize=7)
        ax.use_sticky_edges = False
        ax.margins(y=0.2)
        ax.set_title(str(var)+'\ntime: '+str(timestep), fontsize=10)

        return fig

    def save_figure(self, fig, directory, filename_root,
                    timestep, ext='.png', close=True):
        """Save a figure.

        Parameters
        ----------
        path : :obj:`str`
            The path (and filename without extension) to save the figure to.

        ext : :obj:`str`, optional
            The file extension (default='.png'). This must be supported by the
            active matplotlib backend (see matplotlib.backends module). Most
            backends support '.png', '.pdf', '.ps', '.eps', and '.svg'. Be
            sure to include the '.' before the extension.

        close : :obj:`bool`, optional
            Whether to close the file after saving.

        Returns
        -------

        """
        if self._save_figs_sequential:
            # save as a padded number with the timestep
            savepath = os.path.join(directory,
                                    filename_root + str(timestep).zfill(5) + ext)
        else:
            # save as "latest"
            savepath = os.path.join(directory,
                                    filename_root + 'latest' + ext)

        fig.savefig(savepath)
        if close:
            plt.close()

    def save_grids(self, var_name, var, ts):
        """Save a grid into an existing netCDF file.

        File should already be open (by :meth:`init_output_grid`) as
        ``self.output_netcdf``.

        Parameters
        ----------
        var_name : :obj:`str`
            The name of the variable to be saved

        var : :obj:`ndarray`
            The numpy array to be saved.

        ts : :obj:`int`
            The current timestep (+1, so human readable)

        Returns
        -------

        """
        _msg = ' '.join(['saving', str(var_name), 'grid'])
        self.log_info(_msg, verbosity=2)
        try:
            self.output_netcdf.variables[var_name][ts, :, :] = var
        except Exception as e:
            _msg = (f'Failed to save {var_name} grid to netCDF file, '
                    f'Exception: {e}')
            self.logger.error(_msg)
            raise Exception(e)

    def save_the_checkpoint(self):
        """Save checkpoint files.

        Saves the grids to a .npz file so that the model can be
        initiated from this point. The timestep of the checkpoint is also
        saved. The values from the model that are saved to the checkpoint.npz
        are the following:

        - Model time
        - Flow velocity and its components
        - Water depth
        - Water stage
        - Topography
        - Surface sand fraction
        - Active layer values
        - Current random seed state

        If `save_checkpoint` is turned on, checkpoints are re-written
        with either a frequency of `checkpoint_dt` or `save_dt` if
        `checkpoint_dt` has not been explicitly defined.
        """
        ckp_file = os.path.join(self.prefix, 'checkpoint.npz')

        # get rng state
        rng_state_list = shared_tools.get_random_state()
        rng_state = np.array(rng_state_list,
                             dtype=object)  # convert to object before saving

        np.savez_compressed(
            ckp_file,
            # time and counter variables
            time=self.time,
            time_iter=self._time_iter,
            save_iter=self._save_iter,
            save_time_since_data=self._save_time_since_data,
            # grids
            eta=self.eta,
            depth=self.depth,
            stage=self.stage,
            uw=self.uw,
            ux=self.ux,
            uy=self.uy,
            qw=self.qw,
            qx=self.qx,
            qy=self.qy,
            sand_frac=self.sand_frac,
            active_layer=self.active_layer,
            # boundary condition / state variables
            H_SL=self._H_SL,
            rng_state=rng_state,
            )



import matplotlib.pyplot as plt
import numpy as np


import _base


fig, ax = plt.subplots(2, 1, figsize=(8, 4), sharex=True)
# plt.subplots_adjust(top=0.7)


day = 0
for i in range(4):
    ax[0].axvline(x=day, color='k', ls='-')
    ax[0].fill_between(day + np.arange(365), np.zeros((len(_base.time),)),
                    4 * np.ones((len(_base.time),)),
                    where=_base.flood, alpha=0.4, edgecolor='none')
    ax[0].plot(day + np.arange(365), _base.hydrograph)
    ax[0].text(day + 220, 2.6, 'year ' + str(i))
    day += 365

ax[0].axvline(x=day, color='k', ls='-')
ax[0].set_ylim(0.9, 3.1)
ax[0].set_xlim(0 - 10, day + 10)
ax[0].set_xlabel(r'time $\rightarrow$')
ax[0].set_yticks([1.05, 2.5])
ax[0].set_yticklabels(['base flow', 'bankfull flow'])


nflood = np.sum(_base.flood)
day = 0
for i in range(4):
    ax[1].axvline(x=day, color='k', ls='-')
    ax[1].fill_between(day + np.arange(nflood + 1), np.zeros((nflood + 1,)),
                    4 * np.ones((nflood + 1,)),
                    alpha=0.4, edgecolor='none')
    ax[1].plot(day + np.arange(nflood + 1), 2.5 * np.ones((nflood + 1),))
    day += nflood

ax[1].axvline(x=day, color='k', ls='-')
ax[1].set_ylim(0.9, 3.1)
ax[1].set_xlim(0 - 10, 365 * 4 + 10)
ax[1].set_yticks([])
ax[1].set_xlabel(r'time $\rightarrow$')
ax[1].set_yticks([1.05, 2.5])
ax[1].set_yticklabels(['base flow', 'bankfull flow'])


plt.show()



import matplotlib.pyplot as plt

import pyDeltaRCM

with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(out_dir=output_dir)

delta.show_attribute('cell_type', grid=False)
delta.show_attribute('cell_type', grid=False)
delta.show_ind([3378, 9145, 11568, 514, 13558])
delta.show_ind((42, 94), 'bs')
delta.show_ind([(41, 8), (42, 10)], 'g^')
plt.show()



# unit tests for water_tools.py

import pytest

import numpy as np

import unittest.mock as mock

from pyDeltaRCM.model import DeltaModel
from . import utilities

from pyDeltaRCM import water_tools


class TestWaterRoute:

    def test_route_water(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # mock top-level methods
        _delta.log_info = mock.MagicMock()
        _delta.init_water_iteration = mock.MagicMock()
        _delta.run_water_iteration = mock.MagicMock()
        _delta.compute_free_surface = mock.MagicMock()
        _delta.finalize_water_iteration = mock.MagicMock()

        # run the method
        _delta.route_water()

        # methods called
        assert (_delta.log_info.called is True)
        assert (_delta.init_water_iteration.call_count == _delta._itermax)
        assert (_delta.run_water_iteration.call_count == _delta._itermax)
        assert (_delta.compute_free_surface.call_count == _delta._itermax)
        assert (_delta.finalize_water_iteration.call_count == _delta._itermax)


class TestWaterRoutingWeights:

    def test_get_weight_at_cell(self):
        # prepare necessary ingredients for test
        ind = (0, 4)
        np.random.seed(0)
        stage = np.random.uniform(0.5, 1, 9)
        eta = np.random.uniform(0, 0.85, 9)
        depth = stage - eta
        depth[depth < 0] = 0
        celltype = np.array([-2, -2, -2, 1, 1, -2, 0, 0, 0])
        dry_thresh = 0.1
        gamma = 0.001962
        theta = 1
        weight_sfc = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=np.float64)
        weight_int = np.array([0, 0, 0, 0, 0, 1, 0, 1, 1], dtype=np.float64)

        # get weights back from the test
        wts = water_tools._get_weight_at_cell_water(
            ind, weight_sfc, weight_int, depth,
            celltype, dry_thresh, gamma, theta)
        assert np.all(wts[[0, 1, 2, 5]] == 0)
        assert wts[4] == 0
        assert np.any(wts[[3, 6, 7, 8]] != 0)


class TestInitWaterIteration:

    def test_fields_updated(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)

        delta.log_info = mock.MagicMock()

        # run the method
        delta.init_water_iteration()

        # assertions
        assert np.all(delta.qxn == 0)
        assert np.all(delta.qyn == 0)
        assert np.all(delta.qwn == 0)
        assert np.all(delta.free_surf_flag == 1)  # all parcels begin as valid
        assert np.all(delta.free_surf_walk_inds == 0)
        assert np.all(delta.sfc_visit == 0)
        assert np.all(delta.sfc_sum == 0)

        assert delta.log_info.call_count == 1


class TestCheckForLoops:

    # set up a simple domain
    #   10x10 domain
    #   2 parcels taken six steps
    free_surf_walk_inds = np.array(
            [[4, 15, 25, 36, 46, 55, 66],
             [5, 15, 26, 35, 44, 53, 62]]
             )
    stage = np.ones((10, 10))
    CTR = 4
    L0 = 1

    def test_check_for_loops_no_loops(self):

        # set up inds to go to new locations
        new_inds0 = np.array([76, 73])
        _step = self.free_surf_walk_inds.shape[1] + 1  # 7
        SL = 0
        stage_minus_SL = self.stage - SL

        new_inds, looped = water_tools._check_for_loops(
            self.free_surf_walk_inds, new_inds0.copy(), _step,
            self.L0, self.CTR, stage_minus_SL)

        # new inds should be same as input
        assert np.all(new_inds == new_inds0)
        # no loops
        assert np.all(looped == [0, 0])

    def test_check_for_loops_relocate_not_loop(self):
        """
        Will relocate parcel (repeated ind), but will not mark as a loop,
        because parcel stage is at SL.
        """
        # set up inds[1] to go to a repeat ind
        new_inds0 = np.array([76, 53])
        _step = self.free_surf_walk_inds.shape[1] + 1  # 7
        SL = 1
        stage_minus_SL = self.stage - SL

        new_inds, looped = water_tools._check_for_loops(
            self.free_surf_walk_inds, new_inds0.copy(), _step,
            self.L0, self.CTR, stage_minus_SL)

        # new inds should NOT be same as input
        assert new_inds[0] == new_inds0[0]
        assert new_inds[1] != new_inds0[1]
        # check that the parcel was thrown along vector
        assert new_inds[1] > 80
        # one loop
        assert np.all(looped == [0, 0])

    def test_check_for_loops_relocate_and_loop(self):
        """
        Will relocate parcel (repeated ind), and mark as a loop, because
        parcel stage is not at SL.
        """
        # set up inds[1] to go to a repeat ind
        new_inds0 = np.array([76, 53])
        _step = self.free_surf_walk_inds.shape[1] + 1  # 7
        SL = 0
        stage_minus_SL = self.stage - SL

        new_inds, looped = water_tools._check_for_loops(
            self.free_surf_walk_inds, new_inds0.copy(), _step,
            self.L0, self.CTR, stage_minus_SL)

        # new inds should NOT be same as input
        assert new_inds[0] == new_inds0[0]
        assert new_inds[1] != new_inds0[1]
        # check that the parcel was thrown along vector
        assert new_inds[1] > 80
        # one loop
        assert np.all(looped == [0, 1])


class TestCalculateNewInds:

    def test_calculate_new_inds(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)

        current_inds = np.array(
            [7754, 7743, 7755], dtype=np.int64)
        new_inds = np.array(
            [6, 1, 4], dtype=np.int64)
        ravel_walk_flat = delta.ravel_walk_flat

        nidx = water_tools._calculate_new_inds(
            current_inds, new_inds,
            ravel_walk_flat)

        nidx_exp = np.array([7754 + delta.eta.shape[1] - 1,
                             7743 + -delta.eta.shape[1],
                             0])
        assert np.all(nidx == nidx_exp)


class TestUpdateQFields:

    def test_update_dirQfield(self):
        # configure necessary ingredients
        np.random.seed(0)
        qx = np.random.uniform(0, 10, 9)
        d = np.array([1, np.sqrt(2), 0])
        astep = np.array([True, True, False])
        inds = np.array([3, 4, 5])
        stepdir = np.array([1, 1, 0])

        # run the method
        qxn = water_tools._update_dirQfield(
            np.copy(qx), d, inds, astep, stepdir)

        # compute the expected value
        qxdiff = qxn - qx
        qxdiff_exp = np.array([1, np.sqrt(2) / 2, 0])

        # assertions
        assert np.all(qxdiff[3:6] == pytest.approx(qxdiff_exp))

    def test_update_absQfield(self):
        # configure necessary ingredients
        np.random.seed(0)
        qw = np.random.uniform(0, 10, 9)
        d = np.array([1, np.sqrt(2), 0])
        astep = np.array([True, True, False])
        inds = np.array([3, 4, 5])
        Qp_water = 0.3
        dx = 1.0

        # run the method
        qwn = water_tools._update_absQfield(
            np.copy(qw), d, inds, astep,
            Qp_water, dx)

        # compute the expected values
        qwdiff = qwn - qw
        diffelem = Qp_water / dx / 2
        qwdiff_exp = np.array([diffelem, diffelem, 0])

        # assertions
        assert np.all(qwdiff[3:6] == pytest.approx(qwdiff_exp))


class TestUpdateFlowField:

    def test_update_flow_field_time0_iteration0(self, tmp_path):
        """
        Check that the flow at the inlet is set as expected
        """
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)

        # mock the log
        delta.log_info = mock.MagicMock()

        # conditions are zero already, but...
        delta._time_iter = 0
        iteration = 0

        # run the method
        delta.update_flow_field(iteration)

        # assertions
        assert np.all(delta.qx[1:, :] == delta.qxn[1:, :])
        assert np.all(delta.qy[1:, :] == delta.qyn[1:, :])

        # check inlet boundary conditon
        assert np.all(delta.qx[0, delta.inlet] == delta.qw0)
        assert np.all(delta.qy[0, delta.inlet] == 0)
        assert np.all(delta.qw[0, delta.inlet] == delta.qw0)

        assert delta.log_info.call_count == 1

    def test_update_flow_field_time1_iteration0(self, tmp_path):
        """
        Check that the flow in domain is set as expected when no flow (qx &
        qy==0)
        """
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)

        delta.log_info = mock.MagicMock()

        # conditions are zero already, but...
        delta._time_iter = 1
        iteration = 0

        # run the method
        qx0 = np.copy(delta.qx)
        delta.update_flow_field(iteration)

        # assertions
        #  not sure what to check here other than that something changed
        assert np.any(delta.qx != qx0)

        # check inlet boundary conditon
        assert np.all(delta.qx[0, delta.inlet] == delta.qw0)
        assert np.all(delta.qy[0, delta.inlet] == 0)
        assert np.all(delta.qw[0, delta.inlet] == delta.qw0)

        assert delta.log_info.call_count == 1

    def test_update_flow_field_time1_iteration1(self, tmp_path):
        """
        Check that the flow in domain is set as expected when no flow (qx &
        qy==0)
        """
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)

        delta.log_info = mock.MagicMock()

        # conditions are zero already, but...
        delta._time_iter = 1
        iteration = 1

        # run the method
        qx0 = np.copy(delta.qx)
        delta.update_flow_field(iteration)

        # assertions
        #  not sure what to check here other than that something changed
        assert np.any(delta.qx != qx0)

        # check inlet boundary conditon
        assert np.all(delta.qx[0, delta.inlet] == delta.qw0)
        assert np.all(delta.qy[0, delta.inlet] == 0)
        assert np.all(delta.qw[0, delta.inlet] == delta.qw0)

        assert delta.log_info.call_count == 1

    def test_update_velocity_field(self, tmp_path):
        """
        Check that flow velocity field is updated as expected
        """
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)

        delta.log_info = mock.MagicMock()

        # run the method
        delta.update_velocity_field()

        # make a mask, which is less restictive that the actual mask
        dmask = delta.depth > 0

        # assertions
        assert np.all(delta.ux[dmask] != 0)
        assert np.all(delta.uw[dmask] != 0)

        assert delta.log_info.call_count == 1



import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import pyDeltaRCM


# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


n = 1
cm = matplotlib.cm.get_cmap('tab10')


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')
_shp = delta.eta.shape


delta.init_water_iteration()
delta.run_water_iteration()


# define a function to fill in the walks of given idx
def _plot_idxs_walks_to_step(delta_inds, _step, _idxs, _ax):
    for i in range(len(_idxs)):
        walk = delta_inds[_idxs[i], :]
        walk = walk[:_step]
        # print(walk)
        pyDeltaRCM.debug_tools.plot_line(
            walk, shape=_shp, color='c',
            multiline=True, nozeros=True)
        yend, xend = pyDeltaRCM.shared_tools.custom_unravel(walk[-1], _shp)
        _ax.plot(xend, yend,
                 marker='o', ms=3, color='c')


# declare the idxs to use:
idxs = np.random.randint(low=0, high=delta._Np_water, size=n)
pidx = 85


# set up axis
# fig, ax = plt.subplots(1, 4, figsize=(10, 5))
fig = plt.figure(constrained_layout=True)
gs = fig.add_gridspec(2, 2)


# fill in axis0
ax0 = fig.add_subplot(gs[0, 0])
pyDeltaRCM.debug_tools.plot_domain(
    delta.eta, ax=ax0, grid=False, cmap='cividis')
_plot_idxs_walks_to_step(
    delta.free_surf_walk_inds, _step=pidx, _idxs=idxs, _ax=ax0)
ax0.set_title('bed elevation')


# fill in axis0
ax1 = fig.add_subplot(gs[0, 1])
pyDeltaRCM.debug_tools.plot_domain(
    delta.stage, ax=ax1, grid=False)
_plot_idxs_walks_to_step(delta.free_surf_walk_inds, _step=pidx, _idxs=idxs, _ax=ax1)
ax1.set_title('water surface (stage)')


ax2 = fig.add_subplot(gs[1, :])
ax2.axhline(y=0, xmin=0, xmax=pidx+1, ls='--', color='0.6')
for i in range(n):
    walk = delta.free_surf_walk_inds[idxs[i], :]
    walk = walk[:pidx]

    ax2.plot(delta.eta.flat[walk], 'k-')
    ax2.plot(delta.stage.flat[walk], '-', color=cm(i))

ax2.set_ylabel('elevation')
ax2.set_xlabel('steps along parcel path')

plt.tight_layout()

plt.show()



import sys
import os
import glob

import numpy as np

import pytest
from pyDeltaRCM.model import DeltaModel
from pyDeltaRCM import shared_tools

# utilities for file writing


def create_temporary_file(tmp_path, file_name):
    d = tmp_path / 'configs'
    d.mkdir(parents=True, exist_ok=True)
    p = d / file_name
    f = open(p, "w")
    return p, f


def write_parameter_to_file(f, varname, varvalue):
    f.write(varname + ': ' + str(varvalue) + '\n')


def write_matrix_to_file(f, keys, lists):
    # assert len(keys) == len(lists)
    f.write('matrix' + ': ' + '\n')
    for i in range(len(keys)):
        f.write('  ' + keys[i] + ': ' + '\n')
        for j in range(len(lists[i])):
            f.write('    ' + '- ' + str(lists[i][j]) + '\n')


def write_set_to_file(f, set_list):
    f.write('set' + ': ' + '\n')
    for i, _set in enumerate(set_list):
        f.write('  - {')
        for j, (k, v) in enumerate(_set.items()):
            f.write(k + ': ' + str(v) + ', ')
        f.write('}' + '\n')


def yaml_from_dict(tmp_path, file_name, _dict=None):
    p, f = create_temporary_file(tmp_path, file_name)
    if (_dict is None):
        _dict = {'out_dir': tmp_path / 'out_dir'}
    elif ('out_dir' not in _dict.keys()):
        _dict['out_dir'] = tmp_path / 'out_dir'

    for k in _dict.keys():
        write_parameter_to_file(f, k, _dict[k])
    f.close()
    return p


@pytest.fixture(scope='function')
def test_DeltaModel(tmp_path):
    file_name = 'user_parameters.yaml'
    p, f = create_temporary_file(tmp_path, file_name)
    write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
    write_parameter_to_file(f, 'Length', 10.0)
    write_parameter_to_file(f, 'Width', 10.0)
    write_parameter_to_file(f, 'seed', 0)
    write_parameter_to_file(f, 'dx', 1.0)
    write_parameter_to_file(f, 'L0_meters', 1.0)
    write_parameter_to_file(f, 'S0', 0.0002)
    write_parameter_to_file(f, 'itermax', 1)
    write_parameter_to_file(f, 'Np_water', 10)
    write_parameter_to_file(f, 'u0', 1.0)
    write_parameter_to_file(f, 'N0_meters', 2.0)
    write_parameter_to_file(f, 'h0', 1.0)
    write_parameter_to_file(f, 'H_SL', 0.0)
    write_parameter_to_file(f, 'SLR', 0.001)
    write_parameter_to_file(f, 'Np_sed', 10)
    write_parameter_to_file(f, 'f_bedload', 0.5)
    write_parameter_to_file(f, 'C0_percent', 0.1)
    write_parameter_to_file(f, 'toggle_subsidence', False)
    write_parameter_to_file(f, 'subsidence_rate', 0.0)
    write_parameter_to_file(f, 'start_subsidence', 50.)
    write_parameter_to_file(f, 'save_eta_figs', False)
    write_parameter_to_file(f, 'save_stage_figs', False)
    write_parameter_to_file(f, 'save_depth_figs', False)
    write_parameter_to_file(f, 'save_discharge_figs', False)
    write_parameter_to_file(f, 'save_velocity_figs', False)
    write_parameter_to_file(f, 'save_eta_grids', False)
    write_parameter_to_file(f, 'save_stage_grids', False)
    write_parameter_to_file(f, 'save_depth_grids', False)
    write_parameter_to_file(f, 'save_discharge_grids', False)
    write_parameter_to_file(f, 'save_velocity_grids', False)
    write_parameter_to_file(f, 'save_dt', 500)
    f.close()
    _delta = DeltaModel(input_file=p)
    return _delta


class FastIteratingDeltaModel:
    """A Fast iterating DeltaModel

    This class is useful in patching the DeltaModel for timing tests. The
    patched DeltaModel uses the random number generation internally, so it
    will verify functionality in any checkpointing scenarios, and overwriting
    only the `solve_water_and_sediment_timestep` method removes most of the jitting compilation
    time and much of the actual computation time.
    """

    def solve_water_and_sediment_timestep(self):
        """PATCH"""

        def _get_random_field(shp):
            """Get a field or randoms using the shared function.

            It is critical to use the `shared_tools.get_random_uniform` for
            reproducibility.
            """
            field = np.zeros(shp, dtype=np.float32)
            for i in range(shp[0]):
                for j in range(shp[1]):
                    field[i, j] = shared_tools.get_random_uniform(1)
            return field

        shp = self.eta.shape
        self.eta += _get_random_field(shp)
        self.uw += _get_random_field(shp)
        self.ux += _get_random_field(shp)
        self.uy += _get_random_field(shp)
        self.depth += _get_random_field(shp)
        self.stage += _get_random_field(shp)


def read_endtime_from_log(log_folder):
    _logs = glob.glob(os.path.join(log_folder, '*.log'))
    assert len(_logs) == 1  # log file exists
    with open(_logs[0], 'r') as _logfile:
        _lines = _logfile.readlines()
        _t = 0
        for i, _line in enumerate(_lines):
            if 'Time: ' in _line:
                _t = _line.split(' ')[6]
                _t = _t.strip(' ;')
        return float(_t)




import numpy as np
from numba import njit
from numba import float32, int64
from numba.experimental import jitclass
from scipy import ndimage
import abc

import warnings

from . import shared_tools

# tools for sediment routing algorithms and deposition/erosion


class sed_tools(abc.ABC):

    def route_sediment(self):
        """Sediment routing main method.

        This is the main method for sediment routing in the model. It is
        called once per `update()` call.

        Internally, this method calls:
            * :obj:`route_all_sand_parcels`
            * :obj:`topo_diffusion`
            * :obj:`route_all_mud_parcels`
        """
        _msg = 'Beginning sediment iteration'
        self.log_info(_msg, verbosity=2)

        # initialize the relevant fields and parcel trackers
        self.hook_init_sediment_iteration()
        self.init_sediment_iteration()

        _msg = 'Beginning sand parcel routing'
        self.log_info(_msg, verbosity=2)
        self.hook_route_all_sand_parcels()
        self.route_all_sand_parcels()

        _msg = 'Beginning topographic diffusion'
        self.log_info(_msg, verbosity=2)
        self.hook_topo_diffusion()
        self.topo_diffusion()

        _msg = 'Beginning mud parcel routing'
        self.log_info(_msg, verbosity=2)
        self.hook_route_all_mud_parcels()
        self.route_all_mud_parcels()

    def sed_route(self):
        """Deprecated, since v1.3.1. Use :obj:`route_sediment`."""
        _msg = ('`sed_route` and `hook_sed_route` are deprecated and '
                'have been replaced with `route_sediment`. '
                'Running `route_sediment` now, but '
                'this will be removed in future release.')
        self.logger.warning(_msg)
        warnings.warn(UserWarning(_msg))
        self.route_sediment()

    def init_sediment_iteration(self):
        """Init the water iteration routine.

        Clear and pad fields in preparation for iterating parcels.
        """
        _msg = 'Initializing water iteration'
        self.log_info(_msg, verbosity=2)

        # pad with edge on depth
        self.pad_depth = np.pad(self.depth, 1, 'edge')

        # clear sediment flux field and deposit volumes
        self.qs[:] = 0
        self.Vp_dep_sand[:] = 0
        self.Vp_dep_mud[:] = 0

    def get_inlet_weights_sediment(self, **kwargs):
        """Get weight for inlet cells for sediment parcels.

        This method determines the *weights* describing which inlet cells
        sediment parcels are fed into the domain (where the cells are
        `self.inlet`).
        Internally, :obj:`pyDeltaRCM.shared_tools._get_inlet_weights` is
        called, returning a balanced weighting across all inlet cells.

        .. note::

            Reimplement this method in custom subclasses as needed. Function
            is passed a string argument `parcel_type`, with information about
            the parcel being routed for convenience.
        """
        return shared_tools.get_inlet_weights(self.inlet)

    def route_all_sand_parcels(self):
        """Route sand parcels; topo diffusion.

        This method largely wraps the :obj:`SandRouter`. First, the number of
        parcels and sand fraction (:obj:`~pyDeltaRCM.DeltaModel.f_bedload`)
        are used to determine starting locations for sand parcels. Next, these
        locations are sent to the `SandRouter`, along with many other model
        state variables.

        Finally, variables are unpacked from the `SandRouter` and updated in
        the model fields, where they are later used by the `MudRouter` and the
        water parcel routing.

        Examples
        --------

        The cumulative effect of routing all sand parcels:

        +-------------------------------------------+------------------------------------------------+
        | .. plot:: sed_tools/_initial_bed_state.py | .. plot:: sed_tools/route_all_sand_parcels.py  |
        +-------------------------------------------+------------------------------------------------+
        """
        _msg = 'Determining sand parcel start indicies'
        self.log_info(_msg, verbosity=2)

        num_starts = int(self._Np_sed * self._f_bedload)
        inlet_weights = self.get_inlet_weights_sediment(parcel_type='sand')
        start_indices = shared_tools.get_start_indices(self.inlet,
                                                       inlet_weights,
                                                       num_starts)

        _msg = 'Supplying model state to SandRouter for iteration'
        self.log_info(_msg, verbosity=2)

        self._sr.run(start_indices, self.eta, self.stage, self.depth,
                     self.cell_type, self.uw, self.ux, self.uy,
                     self.Vp_dep_mud, self.Vp_dep_sand,
                     self.qw, self.qx, self.qy, self.qs)

        # These are the variables updated at the end of the `SandRouter`. If
        # you attempt to drop in a replacement SandRouter, you will need to
        # update these fields!!
        _msg = 'Updating DeltaModel based on SandRouter change'
        self.log_info(_msg, verbosity=2)

        self.Vp_dep_mud = self._sr.Vp_dep_mud
        self.Vp_dep_sand = self._sr.Vp_dep_sand
        self.eta = self._sr.eta  # update bed
        self.depth = self._sr.depth  # update depth
        self.pad_depth = self._sr.pad_depth  # update depth in padded array
        self.uw = self._sr.uw  # update absolute flow field
        self.ux = self._sr.ux  # update component flow field
        self.uy = self._sr.uy  # update component flow fielda
        self.qs = self._sr.qs

    def route_all_mud_parcels(self):
        """Route mud parcels.

        This method largely wraps the :obj:`MudRouter`. First, the number of
        parcels and sand fraction (:obj:`~pyDeltaRCM.DeltaModel.f_bedload`)
        are used to determine starting locations for mud parcels. Next, these
        locations are sent to the `MudRouter`, along with many other model
        state variables.

        Finally, variables are unpacked from the `MudRouter` and updated in
        the model fields, where they are later used by the water parcel
        routing.

        Examples
        --------

        The cumulative effect of routing all sand parcels:

        +-------------------------------------------+------------------------------------------------+
        | .. plot:: sed_tools/_initial_bed_state.py | .. plot:: sed_tools/route_all_mud_parcels.py   |
        +-------------------------------------------+------------------------------------------------+
        """
        _msg = 'Determining mud parcel start indicies'
        self.log_info(_msg, verbosity=2)

        num_starts = int(self._Np_sed * (1 - self._f_bedload))
        inlet_weights = self.get_inlet_weights_sediment(parcel_type='mud')
        start_indices = shared_tools.get_start_indices(self.inlet,
                                                       inlet_weights,
                                                       num_starts)

        _msg = 'Supplying model state to MudRouter for iteration'
        self.log_info(_msg, verbosity=2)

        self._mr.run(start_indices, self.eta, self.stage, self.depth,
                     self.cell_type, self.uw, self.ux, self.uy,
                     self.Vp_dep_mud, self.Vp_dep_sand,
                     self.qw, self.qx, self.qy)

        # These are the variables updated at the end of the `MudRouter`. If
        # you attempt to drop in a replacement MudRouter, you will need to
        # update these fields!!
        _msg = 'Updating DeltaModel based on MudRouter change'
        self.log_info(_msg, verbosity=2)

        self.Vp_dep_mud = self._mr.Vp_dep_mud
        self.Vp_dep_sand = self._mr.Vp_dep_sand
        self.eta = self._mr.eta  # update bed
        self.depth = self._mr.depth  # update depth
        self.pad_depth = self._mr.pad_depth  # update depth in padded array
        self.uw = self._mr.uw  # update absolute flow field
        self.ux = self._mr.ux  # update component flow field
        self.uy = self._mr.uy  # update component flow field

    def topo_diffusion(self):
        """Diffuse topography after routing.

        Diffuse topography after routing all coarse sediment parcels. The
        operation is repeated `N_crossdiff` times.
        """
        for _ in range(self.N_crossdiff):

            a = ndimage.convolve(self.eta, self.kernel1, mode='constant')
            b = ndimage.convolve(self.qs, self.kernel2, mode='constant')
            c = ndimage.convolve(self.qs * self.eta, self.kernel2,
                                 mode='constant')

            self.cf = (self.diffusion_multiplier *
                       (self.qs * a - self.eta * b + c))

            self.cf[self.cell_type == -2] = 0
            self.cf[0, :] = 0

            self.eta += self.cf


@njit
def _get_weight_at_cell_sediment(ind, weight_int, depth_nbrs, ct_nbrs,
                                 dry_depth, theta, distances_flat):
    """Get neighbor weight array for sediment routing.

    .. todo::

        Expand description. Include example? Equation? Link to Morphodynamics
        document.
    """
    dry = (depth_nbrs <= dry_depth)
    wall = (ct_nbrs == -2)
    ctr = (np.arange(9) == 4)
    drywall = np.logical_or(dry, wall)

    # always set ctr to 0 before rebalancing
    weight_int[ctr] = 0

    weight = np.copy(weight_int)  # no gamma weighting here
    weight = (depth_nbrs ** theta) * weight

    # ALWAYS disallow the choice to not move
    weight[ctr] = 0

    # ALWAYS disallow the choice to go into land or dry cells
    weight[drywall] = 0

    # sanity check
    if np.any(np.isnan(weight)):
        raise RuntimeError('NaN encountered in sediment weighting.'
                           'Please report error.')

    # correct the weights for random choice
    if np.sum(weight) == 0:
        # if no weights
        #  convert to random walk into any non-wall,
        #  (incl dry cells), weighted by distance
        weight[:] = 1 / distances_flat

        # disallow movement into walls
        weight[wall] = 0

    weight[ctr] = 0  # enforce

    # sanity check
    weight_sum = np.sum(weight)
    if weight_sum == 0:
        raise RuntimeError('No weights encountered in sediment weighting.'
                           'Please report error.')

    # final rebalance
    weight = weight / weight_sum

    return weight


r_spec = [('_dt', float32), ('_dx', float32),
          ('num_starts', int64), ('start_indices', int64[:]),
          ('stepmax', float32), ('px', int64), ('py', int64),
          ('eta', float32[:, :]), ('stage', float32[:, :]),
          ('depth', float32[:, :]), ('cell_type', int64[:, :]),
          ('uw', float32[:, :]), ('ux', float32[:, :]), ('uy', float32[:, :]),
          ('pad_stage', float32[:, :]), ('pad_depth', float32[:, :]),
          ('pad_cell_type', int64[:, :]), ('qw', float32[:, :]),
          ('qx', float32[:, :]), ('qy', float32[:, :]), ('qs', float32[:, :]),
          ('ivec_flat', float32[:]), ('jvec_flat', float32[:]),
          ('iwalk_flat', int64[:]), ('jwalk_flat', int64[:]),
          ('distances_flat', float32[:]),
          ('dry_depth', float32), ('_lambda', float32),
          ('_beta', float32),  ('_f_bedload', float32),
          ('theta_sed', float32), ('u_max', float32),
          ('qs0', float32), ('_u0', float32), ('Vp_sed', float32),
          ('Vp_res', float32), ('Vp_dep_mud', float32[:, :]),
          ('Vp_dep_sand', float32[:, :]),
          ('U_dep_mud', float32), ('U_ero_mud', float32),
          ('U_ero_sand', float32)]


class BaseRouter(object):
    """BaseRouter.

    Defines common methods for jitted routers.

    Subclasses need to define `run`, `_route_one_parcel`, and
    `_deposit_or_erode`.

    .. note::

        Although this class is configured with `@abc.abstractmethod` methods,
        we cannot actually enforce the implementation of this `BaseRouter` as
        an abstract class, because Numba does not support abstract classes
        (yet; https://github.com/numba/numba/issues/6033). The classes are
        included in the base class with abstract decorators to help make it
        clear that these methods need to be implemented in subclassing
        `Router`.
    """
    @abc.abstractmethod
    def run(self, *args, **kwargs):
        ...

    @abc.abstractmethod
    def _route_one_parcel(self):
        ...

    def _choose_next_location(self, px, py):

        # choose next location with weights
        stage_nbrs = self.pad_stage[
            px:px + 3, py:py + 3]
        depth_nbrs = self.pad_depth[
            px:px + 3, py:py + 3]
        cell_type_ind = self.pad_cell_type[
            px:px + 3, py:py + 3]

        _, weight_int = shared_tools.get_weight_sfc_int(
            self.stage[px, py], stage_nbrs.ravel(), self.qx[px, py],
            self.qy[px, py], self.ivec_flat, self.jvec_flat,
            self.distances_flat)

        if np.any(np.isnan(weight_int)):
            raise RuntimeError('NaN in weight_int.')

        if not np.all(np.isfinite(depth_nbrs.ravel())):
            raise RuntimeError('nonfinite in depth_nbrs.')

        weights = _get_weight_at_cell_sediment(
            (px, py), weight_int, depth_nbrs.ravel(),
            cell_type_ind.ravel(), self.dry_depth, self.theta_sed, self.distances_flat)

        new_cell = shared_tools.random_pick(weights)

        dist, istep, jstep, _ = shared_tools.get_steps(
            new_cell, self.iwalk_flat, self.jwalk_flat)

        return istep, jstep, dist

    @abc.abstractmethod
    def _deposit_or_erode(self, px, py):
        """Determine whether to erode or deposit.

        This is the decision making component of the routine, and will be
        different for sand vs mud.
        """
        ...

    def _update_fields(self, Vp_change, px, py):
        """Execute deposit of sand or mud.

        Deposit sediment volume `Vp_change`. The change in bed elevation
        depends on the sediment mass conservation (i.e., Exner equation) and
        is equal to:

        .. code::

            Vp_change / (dx * dx)

        Following the sediment deposition/erosion, the new values for flow
        depth and flow velocity fields are determined.

        .. note::

            Total sediment mass is preserved, but individual categories
            of sand and mud are not. I.e., it is assumed that there is infinite
            sand and/or mud to erode at any location where erosion is
            occurring.

        Parameters
        ----------
        Vp_change : :obj:`float`
            Volume of sediment to deposit / erode. If erosion, `Vp_change`
            should be negative.

        px : :obj:`int`
            Index.

        py : :obj:`int`
            Index.

        Returns
        -------
        """
        # Compute the changes in the bed elevation, and some flow fields
        #    The :obj:`_update_helper` function is a jitted routine which
        #    computes the new values for the bed elevation and then determines
        #    the updated values of the depth and flow velocity fields. These
        #    determinations require several comparisons and repeated indexing,
        #    so we use a jitted "helper" function to do the operations.
        qw0 = self.qw[px, py]
        eta_change = Vp_change / (self._dx * self._dx)

        eta = self.eta[px, py] + eta_change  # new bed
        depth = self.stage[px, py] - eta  # new depth
        depth = np.maximum(0, depth)  # force >=0

        if depth > 0:
            uw = np.minimum(self.u_max, qw0 / depth)
        else:
            uw = self.uw[px, py]

        # now apply the computed updated values
        self.eta[px, py] = eta  # update bed
        self.depth[px, py] = depth  # update depth
        self.pad_depth[px + 1, py + 1] = depth  # update depth in padded array
        self.uw[px, py] = uw  # update absolute flow field
        # update component flow fields
        if qw0 > 0:
            self.ux[px, py] = uw * self.qx[px, py] / qw0
            self.uy[px, py] = uw * self.qy[px, py] / qw0
        else:
            self.ux[px, py] = 0
            self.uy[px, py] = 0

    def _compute_Vp_ero(self, Vp_sed, U_loc, U_ero, beta):
        """Compute volume erorded based on velocity.

        The volume of sediment eroded depends on the local flow velocity
        (:obj:`U_loc`), the critical velocity for erosion (:obj:`U_ero`), and
        `beta`.

        Function is used by both sand and mud pathways, but the input value of
        U_ero depends on the pathway.

        Parameters
        ----------
        Vp_sed : :obj:`float`
            Volume of a sediment parcel.

        U_loc : :obj:`float`
            Local flow velocity.

        U_ero : :obj:`float`
            Critical velocity for erosion.

        beta : :obj:`float`
            unknown.

        Returns
        -------
        Vp_ero : :obj:`float`
            Volume of eroded sediment.
        """
        return (Vp_sed * (U_loc**beta - U_ero**beta) /
                U_ero**beta)

    def _limit_Vp_change(self, Vp, stage, eta, dx, dep_ero):
        """Limit change in volume to 1/4 of a cell volume.

        Function is used by multiple pathways in `mud_dep_ero` and `sand_dep_ero`
        but with different inputs for the sediment volume (:obj:`Vp`).

        dep_ero indicates which pathway we are in, dep==0, ero==1
        """
        depth = stage - eta
        if depth < 0:
            if dep_ero == 0:
                return 0
            else:
                fourth = np.abs(depth) / 4 * (dx * dx)
                return np.minimum(Vp, fourth)
        else:
            fourth = depth / 4 * (dx * dx)
            return np.minimum(Vp, fourth)


@jitclass(r_spec)
class SandRouter(BaseRouter):
    """Jitted class to route sand.

    Initialized in
    :obj:`~pyDeltaRCM.init_tools.init_tools.init_sediment_routers` with a
    multitude of constants. The `Router` is then called via the public
    :obj:`run` method in the :obj:`~pyDeltaRCM.sed_tools.sed_tools.sed_route`
    method.

    .. important::

        A new Router must be initialized if any constants are changed to the
        underlying `DeltaModel` object. Convention is to call
        :obj:`~pyDeltaRCM.init_tools.init_tools.init_sediment_routers` in any
        property `setter`.
    """
    def __init__(self, _dt, dx, Vp_sed, u_max, qs0, u0, U_ero_sand, f_bedload,
                 ivec_flat, jvec_flat, iwalk_flat, jwalk_flat, distances_flat,
                 dry_depth, beta, stepmax, theta_sed):

        self._dt = _dt
        self._dx = dx
        self.Vp_sed = Vp_sed

        self.u_max = u_max
        self.qs0 = qs0
        self._u0 = u0
        self.U_ero_sand = U_ero_sand
        self._f_bedload = f_bedload

        self.ivec_flat, self.jvec_flat,  = ivec_flat, jvec_flat
        self.iwalk_flat, self.jwalk_flat = iwalk_flat, jwalk_flat
        self.distances_flat = distances_flat

        self.dry_depth = dry_depth
        self._beta = beta
        self.stepmax = stepmax
        self.theta_sed = theta_sed

    def run(self, start_indices, eta, stage, depth, cell_type,
            uw, ux, uy, Vp_dep_mud, Vp_dep_sand,
            qw, qx, qy, qs):
        """The main function to route and deposit/erode sand parcels.

        Algorithm is to:

            1. as input, receive the current status of fields from the model.
            Additionally, receive a list of the starting points to use to run
            parcels, as :obj:`px` and :obj:`py`.

            2. begin a `for` loop to run each parcel in series.

            3. in the :obj:`SandRouter`, the sediment partitioning from a
            ghost node is executed. This step is skipped for
            :obj:`MudRouter`.

            4. call :obj:`_route_one_parcel` method to run a single parcel of
            sediment through all iterations.

            5. repeat from 3, until the correct number of parcels have been
            routed. Note that the number of *total* parcels is
            :obj:`~pyDeltaRCM.DeltaModel.Np_sed`, and the number of sand or
            mud parcels will depend on the value of
            :obj:`~pyDeltaRCM.DeltaModel.f_bedload`.

        .. note::

            We are unable to precompute the routing weights, in the way
            we do in :obj:`~pyDeltaRCM.water_tools.get_water_weight_array`,
            because the weighting changes with each parcel step (i.e.,
            morphodynamics).
        """
        self.eta = eta
        self.stage = stage
        self.depth = depth
        self.cell_type = cell_type
        self.uw = uw
        self.ux = ux
        self.uy = uy
        self.pad_stage = shared_tools.custom_pad(stage)
        self.pad_depth = shared_tools.custom_pad(depth)
        self.pad_cell_type = shared_tools.custom_pad(cell_type)
        self.Vp_dep_mud = Vp_dep_mud
        self.Vp_dep_sand = Vp_dep_sand
        self.qw = qw
        self.qx = qx
        self.qy = qy
        self.qs = qs

        num_starts = start_indices.shape[0]
        for np_sed in range(num_starts):

            self.Vp_res = self.Vp_sed

            px = 0
            py = start_indices[np_sed]

            self.qs[px, py] = (self.qs[px, py] +
                               self.Vp_res / 2. / self._dt / self._dx)
            self._route_one_parcel(px, py)

    def _route_one_parcel(self, px, py):
        """Route one parcel.

        Algorithm is to:

            1. as input, receive the starting points to use to run a single
            parcel, as :obj:`px` and :obj:`py`.

            2. begin a while loop, which counts the number of steps the parcel
            has taken

            3. determine, based on the water surface and flow velocity field,
            which location to travel to next. This determination is named as
            :obj:`choose_next_location`, which utilizes the `shared_tools`
            methods to pick the next location.

            4. call the :obj:`_deposit_or_erode` method to detemine whether to
            deposit or erode sediment. This method is implemented
            *differently* for sand and mud routing, and depends on a multitude
            of different model variables. This step modifies the bed elevation
            and flow depth fields (in subfunction :obj:`_update_fields`),
            which necessitates finding the new weights for routing on each
            step. Also, the volume of sediment is either increased or
            decreased (erosion or deposition).

            5. repeat from 3, until `stepmax` is reached, or an "edge" cell is
            reached.
        """
        it = 0
        sed_continue = True

        while sed_continue:

            px0 = px
            py0 = py

            # Choose the next location for the parcel to travel
            istep, jstep, dist = self._choose_next_location(px0, py0)

            px = (px0 + jstep)
            py = (py0 + istep)

            self._partition_sediment(px0, py0, px, py, dist)
            self._deposit_or_erode(px, py)

            it += 1
            if self.cell_type[px, py] == -1:  # check for "edge" cell
                sed_continue = False  # kill the `while` loop
            if (it == self.stepmax):
                sed_continue = False

    def _partition_sediment(self, px0, py0, px, py, dist):
        """Spread sediment flux between two cells.
        """
        partition = self.Vp_res / 2. / self._dt / self._dx
        if dist > 0:
            self.qs[px0, py0] += partition  # deposition in current cell
            self.qs[px, py] += partition  # deposition in new cell

    def _deposit_or_erode(self, px, py):
        """Decide if deposit or erode sand.

        .. note:: Volumetric change is limited to 1/4 local cell water volume.

        Sand deposition:
            If more sediment is in transport (`qs_loc`) than the determined
            transport capacity of the cell (`qs_cap`), sediment needs to
            deposit on the bed.
        Sand erosion:
            Can only occur if local velocity (`U_loc`) is greater than the
            critical erosion threshold for sand
            (:obj:`~pyDeltaRCM.DeltaModel.U_ero_sand`), *and* if the local
            transport capacity is not yet reached (`qs_loc < qs_cap`).
        """
        U_loc = self.uw[px, py]
        qs_cap = (self.qs0 * self._f_bedload / self._u0**self._beta *
                  U_loc**self._beta)
        qs_loc = self.qs[px, py]

        Vp_change = 0
        if qs_loc > qs_cap:
            # Sand deposition
            #     If more sediment is in transport than the determined
            #     transport capacity of the cell (`qs_cap`), sediment needs to
            #     deposit on the bed.
            Vp_change = self._limit_Vp_change(self.Vp_res, self.stage[px, py],
                                              self.eta[px, py], self._dx, 0)

        elif (U_loc > self.U_ero_sand) and (qs_loc < qs_cap):
            # Sand erosion
            #     Can only occur if local velocity is greater than the
            #     critical erosion threshold for sand, *and* if the local
            #     transport capacity is not yet reached.
            Vp_change = self._compute_Vp_ero(self.Vp_sed, U_loc,
                                             self.U_ero_sand, self._beta)
            Vp_change = self._limit_Vp_change(Vp_change, self.stage[px, py],
                                              self.eta[px, py], self._dx, 1)
            Vp_change = Vp_change * -1

        if Vp_change > 0:  # if deposition
            self.Vp_dep_sand[px, py] = self.Vp_dep_sand[px, py] + Vp_change

        self.Vp_res = self.Vp_res - Vp_change  # update sed volume in parcel

        self._update_fields(Vp_change, px, py)  # update other fields as needed


@jitclass(r_spec)
class MudRouter(BaseRouter):
    """Jitted class to route mud.

    Initialized in
    :obj:`~pyDeltaRCM.init_tools.init_tools.init_sediment_routers` with a
    multitude of constants. The `Router` is then called via the public
    :obj:`run` method in the :obj:`~pyDeltaRCM.sed_tools.sed_tools.sed_route`
    method.

    .. important::

        A new Router must be initialized if any constants are changed to the
        underlying `DeltaModel` object. Convention is to call
        :obj:`~pyDeltaRCM.init_tools.init_tools.init_sediment_routers` in any
        property `setter`.
    """
    def __init__(self, _dt, dx, Vp_sed, u_max, U_dep_mud, U_ero_mud,
                 ivec_flat, jvec_flat, iwalk_flat, jwalk_flat, distances_flat,
                 dry_depth, _lambda, beta, stepmax, theta_sed):

        self._dt = _dt
        self._dx = dx
        self.Vp_sed = Vp_sed

        self.u_max = u_max
        self.U_dep_mud = U_dep_mud
        self.U_ero_mud = U_ero_mud

        self.ivec_flat, self.jvec_flat,  = ivec_flat, jvec_flat
        self.iwalk_flat, self.jwalk_flat = iwalk_flat, jwalk_flat
        self.distances_flat = distances_flat

        self.dry_depth = dry_depth
        self._lambda = _lambda
        self._beta = beta
        self.stepmax = stepmax
        self.theta_sed = theta_sed

    def run(self, start_indices, eta, stage, depth, cell_type,
            uw, ux, uy, Vp_dep_mud, Vp_dep_sand,
            qw, qx, qy):
        """The main function to route and deposit/erode mud parcels.

        """

        self.eta = eta
        self.stage = stage
        self.depth = depth
        self.cell_type = cell_type
        self.uw = uw
        self.ux = ux
        self.uy = uy
        self.pad_stage = shared_tools.custom_pad(stage)
        self.pad_depth = shared_tools.custom_pad(depth)
        self.pad_cell_type = shared_tools.custom_pad(cell_type)
        self.Vp_dep_mud = Vp_dep_mud
        self.Vp_dep_sand = Vp_dep_sand
        self.qw = qw
        self.qx = qx
        self.qy = qy

        num_starts = start_indices.shape[0]
        for np_sed in range(num_starts):

            self.Vp_res = self.Vp_sed

            px = 0
            py = start_indices[np_sed]

            self._route_one_parcel(px, py)

    def _route_one_parcel(self, px, py):
        """Route one parcel.

        """
        it = 0
        sed_continue = True

        while sed_continue:

            # Choose the next location for the parcel to travel
            istep, jstep, _ = self._choose_next_location(px, py)

            px = px + jstep
            py = py + istep

            self._deposit_or_erode(px, py)

            it += 1
            if self.cell_type[px, py] == -1:  # check for "edge" cell
                sed_continue = False  # kill the `while` loop
            if (it == self.stepmax):
                sed_continue = False

    def _deposit_or_erode(self, px, py):
        """Decide if deposit or erode mud.

        .. note:: Volumetric change is limited to 1/4 local cell water volume.

        .. important:: TODO: complete description specific for mud transport
        """
        U_loc = self.uw[px, py]

        Vp_change = 0
        if U_loc < self.U_dep_mud:
            Vp_change = (self._lambda * self.Vp_res *
                         (self.U_dep_mud**self._beta - U_loc**self._beta) /
                         (self.U_dep_mud**self._beta))
            Vp_change = self._limit_Vp_change(Vp_change, self.stage[px, py],
                                              self.eta[px, py], self._dx, 0)

        if U_loc > self.U_ero_mud:
            Vp_change = self._compute_Vp_ero(self.Vp_sed, U_loc,
                                             self.U_ero_mud, self._beta)
            Vp_change = self._limit_Vp_change(Vp_change, self.stage[px, py],
                                              self.eta[px, py], self._dx, 1)
            Vp_change = Vp_change * -1

        if Vp_change > 0:  # if deposition
            self.Vp_dep_mud[px, py] = self.Vp_dep_mud[px, py] + Vp_change

        self.Vp_res = self.Vp_res - Vp_change  # update sed volume in parcel

        self._update_fields(Vp_change, px, py)  # update other fields as needed



# unit tests for shared_tools.py

import pytest

import numpy as np

from pyDeltaRCM.model import DeltaModel
from pyDeltaRCM import shared_tools
from . import utilities


class TestGetAndSetRandom:

    def test_set_random_get_expected(self, tmp_path):
        """
        Test for function shared_tools.get_random_uniform and
        test for function shared_tools.set_random_seed
        """
        # set the random seed
        shared_tools.set_random_seed(0)

        # get back an expected value
        got = shared_tools.get_random_uniform(1)

        # assertion against expected value
        _exp = 0.5488135039273248
        assert got == pytest.approx(_exp)


class TestGetSteps:

    def test_get_steps(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)

        new_cells = np.arange(9)
        iwalk = delta.iwalk
        jwalk = delta.jwalk

        d, i, j, a = shared_tools.get_steps(
            new_cells, iwalk.flatten(), jwalk.flatten())

        d_exp = np.array([1.41421356, 1., 1.41421356, 1., 0.,
                          1., 1.41421356, 1., 1.41421356])
        i_exp = np.array([-1,  0,  1, -1,  0,  1, -1,  0,  1])
        j_exp = np.array([-1, -1, -1,  0,  0,  0,  1,  1,  1])

        # assertions
        assert np.all(np.delete(a, 4))
        assert ~a[4]
        assert i == pytest.approx(i_exp)
        assert j == pytest.approx(j_exp)
        assert d == pytest.approx(d_exp)


class TestRandomPick:

    def test_random_pick(self):
        """
        Test for function shared_tools.random_pick
        """
        # define probs array of zeros with a single 1 value
        probs = np.zeros((9,))
        probs[0] = 1
        # should return first index
        assert shared_tools.random_pick(probs) == 0

    def test_random_pick_diff_fixed(self):
        """
        Test for function shared_tools.random_pick
        """
        # define probs array of zeros with a single 1 value
        probs = np.zeros((9,))
        probs[2] = 1
        # should return third index
        assert shared_tools.random_pick(probs) == 2

    def test_random_pick_not_zeroed(self):
        """
        Test for function shared_tools.random_pick
        """
        # define probs array of zeros with a single 1 value
        probs = np.zeros((9,))
        probs[:] = 1/6
        probs[4] = 0
        probs[5] = 0
        probs[8] = 0
        assert np.sum(probs) == 1
        # should return first index
        _rets = np.zeros((100,))
        for i in range(100):
            _rets[i] = shared_tools.random_pick(probs)
        assert np.all(_rets != 4)
        assert np.all(_rets != 5)
        assert np.all(_rets != 8)

    def test_random_pick_distribution(self):
        """
        Test for function shared_tools.random_pick
        """
        # define probs array of zeros with a single 1 value
        probs = np.array([1/6, 1/3, 0,
                          1/12, 1/12, 1/6,
                          0, 0, 1/6])
        assert np.sum(probs) == 1
        _rets = np.zeros((10000,))
        for i in range(10000):
            _rets[i] = shared_tools.random_pick(probs)
        _bins = np.arange(-0.5, 9, step=1.0)
        _hist, _ = np.histogram(_rets, bins=_bins)
        _binedge = _bins[:-1]
        _histnorm = _hist / np.sum(_hist)

        assert np.all(_histnorm == pytest.approx(probs, rel=0.10))

    def test_random_pick_anybut_first(self, tmp_path):
        """
        Test for function shared_tools.random_pick
        """
        probs = (1 / 7) * np.ones((3, 3), dtype=np.float64)
        probs[0, 0] = 0
        probs[1, 1] = 0
        probs_flat = probs.flatten()

        assert np.sum(probs_flat) == 1
        # should never return first index
        _rets = np.zeros((100,))
        for i in range(100):
            _rets[i] = shared_tools.random_pick(probs_flat)

        assert np.all(_rets != 0)
        assert np.all(_rets != 4)  # THIS LINE NEEDS TO PASS!!
        assert np.sum(_rets == 1) > 0
        assert np.sum(_rets == 2) > 0
        assert np.sum(_rets == 3) > 0
        assert np.sum(_rets == 5) > 0
        assert np.sum(_rets == 6) > 0
        assert np.sum(_rets == 7) > 0
        assert np.sum(_rets == 8) > 0


class TestCustomUnravel:

    def test_custom_unravel_square(self):
        arr = np.arange(9).reshape((3, 3))
        # test upper left corner
        x, y = shared_tools.custom_unravel(0, arr.shape)
        assert x == 0
        assert y == 0
        # test center
        x, y = shared_tools.custom_unravel(4, arr.shape)
        assert x == 1
        assert y == 1
        # test off-center
        x, y = shared_tools.custom_unravel(5, arr.shape)
        assert x == 1
        assert y == 2
        # test lower right corner
        x, y = shared_tools.custom_unravel(8, arr.shape)
        assert x == 2
        assert y == 2

    def test_custom_unravel_rectangle(self):
        arr = np.arange(50).reshape((5, 10))
        # test a few spots
        x, y = shared_tools.custom_unravel(19, arr.shape)
        assert x == 1
        assert y == 9
        x, y = shared_tools.custom_unravel(34, arr.shape)
        assert x == 3
        assert y == 4

    def test_custom_unravel_exceed_error(self):
        arr = np.arange(9).reshape((3, 3))
        # next line should throw IndexError
        with pytest.raises(IndexError):
            x, y = shared_tools.custom_unravel(99, arr.shape)


class TestCustomRavel:

    def test_custom_ravel_square(self):
        arr = np.arange(9).reshape((3, 3))
        # test upper left corner
        tup = (0, 0)
        i = shared_tools.custom_ravel(tup, arr.shape)
        assert i == 0
        # test center
        tup = (1, 1)
        i = shared_tools.custom_ravel(tup, arr.shape)
        assert i == 4
        # test off-center
        tup = (1, 2)
        i = shared_tools.custom_ravel(tup, arr.shape)
        assert i == 5
        # test lower right corner
        tup = (2, 2)
        i = shared_tools.custom_ravel(tup, arr.shape)
        assert i == 8

    def test_custom_ravel_rectangle(self):
        arr = np.arange(50).reshape((5, 10))
        # test a few spots
        tup = (1, 9)
        i = shared_tools.custom_ravel(tup, arr.shape)
        assert i == 19
        tup = (3, 4)
        i = shared_tools.custom_ravel(tup, arr.shape)
        assert i == 34

    def test_custom_ravel_exceed_error(self):
        arr = np.arange(9).reshape((3, 3))
        # next line should throw IndexError
        with pytest.raises(IndexError):
            _ = shared_tools.custom_ravel((50, 50), arr.shape)


class TestGetWeightSfcInt:

    def test_get_weight_sfc_int(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        delta = DeltaModel(input_file=p)

        stage = np.random.uniform(0.5, 1, 9)
        qx = 1
        qy = 1
        ivec = delta.ivec_flat
        jvec = delta.jvec_flat
        dists = delta.distances_flat

        weight_sfc, weight_int = shared_tools.get_weight_sfc_int(
            stage[4], stage, qx, qy, ivec, jvec, dists)
        # assert np.all(weight_sfc == np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]))
        # assert np.all(weight_int == np.array([0, 0, 0, 0, 0, 1, 0, 1, 1]))
        assert np.all(weight_sfc <= 1)
        assert np.all(weight_int <= 1)
        assert np.all(weight_sfc >= 0)
        assert np.all(weight_int >= 0)


class TestVerisonSpecifications:

    def test_version_is_valid(self):
        v = shared_tools._get_version()
        assert type(v) is str
        dots = [i for i, c in enumerate(v) if c == '.']
        assert len(dots) == 2


class TestScaleModelTime():

    def test_defaults(self):
        scaled = shared_tools.scale_model_time(86400)
        assert scaled == 86400

    def test_If(self):
        scaled = shared_tools.scale_model_time(86400, 0.1)
        assert scaled == 86400 / 0.1
        scaled = shared_tools.scale_model_time(86400, 0.5)
        assert scaled == 86400 / 0.5
        scaled = shared_tools.scale_model_time(86400, 0.9)
        assert scaled == 86400 / 0.9
        scaled = shared_tools.scale_model_time(86400, 1)
        assert scaled == 86400
        scaled = shared_tools.scale_model_time(86400, 1e-15)
        assert scaled == 86400 / 1e-15
        with pytest.raises(ValueError, match='Intermittency `If` .*'):
            scaled = shared_tools.scale_model_time(86400, 0)
        with pytest.raises(ValueError, match='Intermittency `If` .*'):
            scaled = shared_tools.scale_model_time(86400, 1.01)

    def test_units(self):
        scaled = shared_tools.scale_model_time(86400, units='seconds')
        assert scaled == 86400
        scaled = shared_tools.scale_model_time(86400, units='days')
        assert scaled == 1
        scaled = shared_tools.scale_model_time(86400, units='years')
        assert scaled == (1 / 365.25)
        with pytest.raises(ValueError):
            scaled = shared_tools.scale_model_time(86400, units='badstr')

    def test_combinations(self):
        scaled = shared_tools.scale_model_time(86400, If=0.1, units='days')
        assert scaled == 10
        scaled = shared_tools.scale_model_time(2 * 86400, If=0.1, units='days')
        assert scaled == 20
        scaled = shared_tools.scale_model_time(
            365.25 * 86400, If=1, units='years')
        assert scaled == 1
        scaled = shared_tools.scale_model_time(
            365.25 * 86400, If=0.1, units='years')
        assert scaled == 10



import numpy as np

time = np.arange(365)
hydrograph = np.ones((len(time),)) + \
    (np.random.uniform(size=(len(time),)) / 10)
hydrograph[100:150] = 2 + np.sin(np.arange(15, 65) / 15)
flood = hydrograph >= 2.5
